# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation.  All rights reserved.
# Licensed under the MIT License.  See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------


import numpy as np
import onnx
from bert_model_generator import float_tensor
from onnx import TensorProto, helper, numpy_helper


# Adapted from bert_model_generator.py
def get_tensor_and_weight(name: str, shape: list[int], random=False, zeros=False):
    low = 0.0
    high = 1.0
    total_elements = 1
    for x in shape:
        total_elements *= x
    weights = (
        [np.random.uniform(low, high) for _ in range(total_elements)]
        if random
        else [0.0] * total_elements
        if zeros
        else [1.0] * total_elements
    )
    return helper.make_tensor(name, TensorProto.FLOAT, shape, weights), weights


def create_whisper_encoder_attention(
    hidden_size=768,
    num_heads=12,
    epsilon=0.000009999999747378752,
    add_before_layernorm=False,
    add_k=False,
    fused=False,
):
    # Get head size and ensure head size is an integer
    assert hidden_size % num_heads == 0
    head_size = hidden_size // num_heads

    # Construct input and output nodes
    inputs = [
        helper.make_tensor_value_info("input_0", TensorProto.FLOAT, ["batch_size", 1500, hidden_size]),
    ]
    outputs = [
        helper.make_tensor_value_info("output_0", TensorProto.FLOAT, ["batch_size", 1500, hidden_size]),
        helper.make_tensor_value_info("output_1", TensorProto.FLOAT, ["batch_size", "sequence_length", hidden_size]),
    ]

    nodes = []
    # Create layernorm (Add + LayerNorm or SkipLayerNorm)
    if add_before_layernorm:
        nodes.extend(
            [
                helper.make_node(
                    "Add", ["input_0", "input_0"], ["layernorm_output_to_skiplayernorm"], "add_before_layernorm"
                ),
                helper.make_node(
                    "LayerNormalization",
                    ["layernorm_output_to_skiplayernorm", "layernorm_weight", "layernorm_bias"],
                    ["layernorm_output_to_matmul"],
                    "layernorm",
                    epsilon=epsilon,
                ),
            ]
        )
    else:
        nodes.append(
            helper.make_node(
                "SkipLayerNormalization",
                ["input_0", "input_0", "layernorm_weight", "layernorm_bias"],
                ["layernorm_output_to_matmul", "", "", "layernorm_output_to_skiplayernorm"],
                "skiplayernorm",
                domain="com.microsoft",
                epsilon=epsilon,
            )
        )

    if fused:
        nodes.append(
            helper.make_node(
                "Attention",
                ["layernorm_output_to_matmul", "Attention_0_qkv_weight", "Attention_0_qkv_bias", ""],
                ["attn_output"],
                "Attention_0",
                domain="com.microsoft",
                num_heads=num_heads,
            ),
        )
    else:
        # Create nodes for Q/K/V paths
        q_nodes = [
            helper.make_node(
                "MatMul", ["layernorm_output_to_matmul", "q_weight"], ["q_matmul_output"], "q_path_matmul"
            ),
            helper.make_node("Add", ["q_bias", "q_matmul_output"], ["q_add_output"], "q_path_add"),
            helper.make_node("Mul", ["q_add_output", "q_scale"], ["q_mul_output"], "q_path_mul"),
            helper.make_node("Reshape", ["q_mul_output", "q_bsnh_reshape"], ["q_4d_bsnh"], "q_reshape_to_4d"),
            helper.make_node("Transpose", ["q_4d_bsnh"], ["q_4d_bnsh"], "q_transpose_to_bnsh", perm=[0, 2, 1, 3]),
            helper.make_node(
                "Reshape",
                ["q_4d_bnsh", "q_attn_heads_output"],
                ["q_output_(num_heads*batch_size,seq_len,head_size)"],
                "q_reshape_to_3d",
            ),
        ]
        k_nodes = [
            helper.make_node(
                "MatMul", ["layernorm_output_to_matmul", "k_weight"], ["k_matmul_output"], "k_path_matmul"
            ),
        ]
        if add_k:
            k_nodes.extend(
                [
                    helper.make_node("Add", ["k_bias", "k_matmul_output"], ["k_add_output"], "k_path_add"),
                    helper.make_node("Reshape", ["k_add_output", "bsnh_reshape"], ["k_4d_bsnh"], "k_reshape_to_4d"),
                ]
            )
        else:
            k_nodes.append(
                helper.make_node("Reshape", ["k_matmul_output", "kv_bsnh_reshape"], ["k_4d_bsnh"], "k_reshape_to_4d"),
            )
        k_nodes.extend(
            [
                helper.make_node("Transpose", ["k_4d_bsnh"], ["k_4d_bnsh"], "k_transpose_to_bnsh", perm=[0, 2, 1, 3]),
                helper.make_node(
                    "Reshape",
                    ["k_4d_bnsh", "k_attn_heads_output"],
                    ["k_output_(num_heads*batch_size,seq_len,head_size)"],
                    "k_reshape_to_3d",
                ),
                helper.make_node(
                    "Transpose",
                    ["k_output_(num_heads*batch_size,seq_len,head_size)"],
                    ["k_output_(num_heads*batch_size,head_size,seq_len)"],
                    "k_transpose_last_two_dims",
                    perm=[0, 2, 1],
                ),
            ]
        )
        v_nodes = [
            helper.make_node(
                "MatMul", ["layernorm_output_to_matmul", "v_weight"], ["v_matmul_output"], "v_path_matmul"
            ),
            helper.make_node("Add", ["v_bias", "v_matmul_output"], ["v_add_output"], "v_path_add"),
            helper.make_node("Reshape", ["v_add_output", "kv_bsnh_reshape"], ["v_4d_bsnh"], "v_reshape_to_4d"),
            helper.make_node("Transpose", ["v_4d_bsnh"], ["v_4d_bnsh"], "v_transpose_to_bnsh", perm=[0, 2, 1, 3]),
            helper.make_node(
                "Reshape",
                ["v_4d_bnsh", "v_attn_heads_output"],
                ["v_output_(num_heads*batch_size,seq_len,head_size)"],
                "v_reshape_to_3d",
            ),
        ]
        nodes.extend(q_nodes)
        nodes.extend(k_nodes)
        nodes.extend(v_nodes)

        # Create nodes used with qkv concats, reshapes, and transposes
        nodes.extend(
            [
                helper.make_node("Shape", ["layernorm_output_to_matmul"], ["shape_output"], "shape"),
                helper.make_node("Gather", ["shape_output", "idx_0"], ["gather_0_output"], "gather_0", axis=0),
                helper.make_node(
                    "Mul", ["gather_0_output", "num_heads_int"], ["mul_attn_heads_output"], "mul_num_heads"
                ),
                helper.make_node(
                    "Unsqueeze",
                    ["mul_attn_heads_output", "unsqueeze_axes_input"],
                    ["unsqueeze_attn_heads_output"],
                    "unsqueeze_num_heads",
                ),
                helper.make_node(
                    "Concat",
                    ["unsqueeze_attn_heads_output", "neg_one", "head_size"],
                    ["q_attn_heads_output"],
                    "q_num_heads",
                    axis=0,
                ),
                helper.make_node(
                    "Concat",
                    ["unsqueeze_attn_heads_output", "neg_one", "head_size"],
                    ["k_attn_heads_output"],
                    "k_num_heads",
                    axis=0,
                ),
                helper.make_node(
                    "Concat",
                    ["unsqueeze_attn_heads_output", "neg_one", "head_size"],
                    ["v_attn_heads_output"],
                    "v_num_heads",
                    axis=0,
                ),
                helper.make_node(
                    "Constant",
                    inputs=[],
                    outputs=["q_bsnh_reshape"],
                    value=numpy_helper.from_array(
                        np.array([0, 0, num_heads, head_size], dtype="int64"), name="const_tensor"
                    ),
                ),
                helper.make_node(
                    "Constant",
                    inputs=[],
                    outputs=["kv_bsnh_reshape"],
                    value=numpy_helper.from_array(
                        np.array([0, -1, num_heads, head_size], dtype="int64"), name="const_tensor"
                    ),
                ),
            ]
        )

        # Create nodes used with Q x K' and softmax(Q x K') x V
        nodes.extend(
            [
                helper.make_node("Gather", ["shape_output", "idx_1"], ["gather_1_output"], "gather_1", axis=0),
                helper.make_node(
                    "Unsqueeze", ["gather_0_output", "unsqueeze_axes_input"], ["unsqueeze_0_output"], "unsqueeze_0"
                ),
                helper.make_node(
                    "Unsqueeze", ["gather_1_output", "unsqueeze_axes_input"], ["unsqueeze_1_output"], "unsqueeze_1"
                ),
                helper.make_node(
                    "Concat",
                    ["unsqueeze_0_output", "num_heads", "unsqueeze_1_output", "head_size"],
                    ["bnsh_format"],
                    axis=0,
                ),
                helper.make_node(
                    "Concat", ["unsqueeze_0_output", "unsqueeze_1_output", "hidden_size"], ["bsd_format"], axis=0
                ),
            ]
        )

        # Create nodes for computing softmax(Q x K') x V
        nodes.extend(
            [
                helper.make_node(
                    "MatMul",
                    [
                        "q_output_(num_heads*batch_size,seq_len,head_size)",
                        "k_output_(num_heads*batch_size,head_size,seq_len)",
                    ],
                    ["qk_output_(num_heads*batch_size,seq_len,seq_len)"],
                    "matmul_qk",
                ),
                helper.make_node(
                    "Softmax",
                    ["qk_output_(num_heads*batch_size,seq_len,seq_len)"],
                    ["softmax_output"],
                    "softmax_qk",
                    axis=2,
                ),
                helper.make_node(
                    "MatMul",
                    ["softmax_output", "v_output_(num_heads*batch_size,seq_len,head_size)"],
                    ["qkv_output_(num_heads*batch_size,seq_len,head_size)"],
                    "matmul_qkv",
                ),
                helper.make_node(
                    "Reshape",
                    ["qkv_output_(num_heads*batch_size,seq_len,head_size)", "bnsh_format"],
                    ["qkv_bnsh"],
                    "reshape_qkv_to_bnsh",
                ),
                helper.make_node("Transpose", ["qkv_bnsh"], ["qkv_bsnh"], "transpose_bnsh_to_bsnh", perm=[0, 2, 1, 3]),
                helper.make_node("Reshape", ["qkv_bsnh", "bsd_format"], ["attn_output"], "qkv_bsd"),
            ]
        )

    # Create final nodes to conclude attention
    nodes.append(
        helper.make_node(
            "MatMul",
            ["attn_output", "matmul_after_attn_initializer"],
            ["matmul_after_attn_output"],
            "matmul_after_attn",
        ),
    )
    if not fused:
        next_sln_inputs = [
            "layernorm_output_to_skiplayernorm",
            "add_after_attn_output",
            "layernorm_weight",
            "layernorm_bias",
        ]
        nodes.extend(
            [
                helper.make_node(
                    "Add",
                    ["add_after_attn_initializer", "matmul_after_attn_output"],
                    ["add_after_attn_output"],
                    "add_after_attn",
                ),
                helper.make_node(
                    "SkipLayerNormalization",
                    next_sln_inputs,
                    ["output_0", "", "", "output_1"],
                    "next_skiplayernorm",
                    domain="com.microsoft",
                    epsilon=epsilon,
                ),
            ]
        )
    else:
        next_sln_inputs = [
            "matmul_after_attn_output",
            "layernorm_output_to_skiplayernorm",
            "layernorm_weight",
            "layernorm_bias",
            "add_after_attn_initializer",
        ]
        nodes.append(
            helper.make_node(
                "SkipLayerNormalization",
                next_sln_inputs,
                ["output_0", "", "", "output_1"],
                "SkipLayerNorm_AddBias_0",
                domain="com.microsoft",
                epsilon=epsilon,
            )
        )

    # Create initializers
    q_weight, q_weight_data = get_tensor_and_weight("q_weight", [hidden_size, hidden_size])
    q_bias, q_bias_data = get_tensor_and_weight("q_bias", [hidden_size])
    k_weight, k_weight_data = get_tensor_and_weight("k_weight", [hidden_size, hidden_size])
    k_bias, k_bias_data = get_tensor_and_weight("k_bias", [hidden_size], zeros=(not add_k))
    v_weight, v_weight_data = get_tensor_and_weight("v_weight", [hidden_size, hidden_size])
    v_bias, v_bias_data = get_tensor_and_weight("v_bias", [hidden_size])
    qkv_weight = helper.make_tensor(
        "Attention_0_qkv_weight",
        TensorProto.FLOAT,
        [hidden_size, 3 * hidden_size],
        q_weight_data + k_weight_data + v_weight_data,
    )
    qkv_bias = helper.make_tensor(
        "Attention_0_qkv_bias", TensorProto.FLOAT, [3 * hidden_size], q_bias_data + k_bias_data + v_bias_data
    )
    initializers = [
        float_tensor("layernorm_weight", [hidden_size]),
        float_tensor("layernorm_bias", [hidden_size]),
        float_tensor("matmul_after_attn_initializer", [hidden_size, hidden_size]),
        float_tensor("add_after_attn_initializer", [hidden_size]),
    ]
    if fused:
        initializers.extend([qkv_weight, qkv_bias])
    else:
        initializers.extend(
            [
                numpy_helper.from_array(np.array(num_heads, dtype="int64"), name="num_heads_int"),
                numpy_helper.from_array(np.array([num_heads], dtype="int64"), name="num_heads"),
                numpy_helper.from_array(np.array([head_size], dtype="int64"), name="head_size"),
                numpy_helper.from_array(np.array([hidden_size], dtype="int64"), name="hidden_size"),
                numpy_helper.from_array(np.array(1 / np.sqrt(head_size), dtype="float32"), name="q_scale"),
                numpy_helper.from_array(np.array(0, dtype="int64"), name="idx_0"),
                numpy_helper.from_array(np.array(1, dtype="int64"), name="idx_1"),
                numpy_helper.from_array(np.array([-1], dtype="int64"), name="neg_one"),
                numpy_helper.from_array(np.array([0], dtype="int64"), name="unsqueeze_axes_input"),
            ]
        )
        if add_k:
            initializers.extend([q_weight, q_bias, k_weight, k_bias, v_weight, v_bias])
        else:
            initializers.extend([q_weight, q_bias, k_weight, v_weight, v_bias])

    # Construct graph
    graph = helper.make_graph(
        nodes, "whisper_encoder_attention_graph", inputs, outputs, initializers, doc_string="whisper"
    )
    opsetid = helper.make_opsetid("ai.onnx", min(onnx.defs.onnx_opset_version(), 16))
    return helper.make_model(graph, opset_imports=(opsetid,))


def create_whisper_decoder_attention(
    hidden_size=768,
    num_heads=12,
    epsilon=0.000009999999747378752,
    add_before_layernorm=False,
    add_k=False,
    fused=False,
):
    # Get head size and ensure head size is an integer
    assert hidden_size % num_heads == 0
    head_size = hidden_size // num_heads

    # Construct input and output nodes
    # Dummy inputs are used to prevent the nodes in the path for the decoder attention mask to be fused together
    # before attention is fused
    inputs = [
        helper.make_tensor_value_info("input_0", TensorProto.FLOAT, ["batch_size", 1500, hidden_size]),
        helper.make_tensor_value_info("dummy_input_int64", TensorProto.INT64, ["dummy_input_1d_int64"]),
        helper.make_tensor_value_info("dummy_input_fp32", TensorProto.FLOAT, ["dummy_input_1d_fp32"]),
    ]
    outputs = [
        helper.make_tensor_value_info(
            "present.0.decoder.key", TensorProto.FLOAT, ["batch_size", num_heads, 1500, head_size]
        ),
        helper.make_tensor_value_info(
            "present.0.decoder.value", TensorProto.FLOAT, ["batch_size", num_heads, 1500, head_size]
        ),
        helper.make_tensor_value_info("output_0", TensorProto.FLOAT, ["batch_size", 1500, hidden_size]),
        helper.make_tensor_value_info("output_1", TensorProto.FLOAT, ["batch_size", "sequence_length", hidden_size]),
    ]

    nodes = []
    # Create layernorm (Add + LayerNorm or SkipLayerNorm)
    if add_before_layernorm:
        nodes.extend(
            [
                helper.make_node(
                    "Add", ["input_0", "input_0"], ["layernorm_output_to_skiplayernorm"], "add_before_layernorm"
                ),
                helper.make_node(
                    "LayerNormalization",
                    ["layernorm_output_to_skiplayernorm", "layernorm_weight", "layernorm_bias"],
                    ["layernorm_output_to_matmul"],
                    "layernorm",
                    epsilon=epsilon,
                ),
            ]
        )
    else:
        nodes.append(
            helper.make_node(
                "SkipLayerNormalization",
                ["input_0", "input_0", "layernorm_weight", "layernorm_bias"],
                ["layernorm_output_to_matmul", "", "", "layernorm_output_to_skiplayernorm"],
                "skiplayernorm",
                domain="com.microsoft",
                epsilon=epsilon,
            )
        )

    if fused:
        nodes.extend(
            [
                helper.make_node(
                    "Attention",
                    [
                        "layernorm_output_to_matmul",
                        "Attention_0_qkv_weight",
                        "Attention_0_qkv_bias",
                        "",
                        "",
                        "attention_add_qk",
                    ],
                    ["attn_output", "present_0_decoder"],
                    "Attention_0",
                    domain="com.microsoft",
                    num_heads=num_heads,
                ),
                helper.make_node(
                    "Gather",
                    ["present_0_decoder", "index_0"],
                    ["present.0.decoder.key"],
                    "Gather_0",
                    axis=0,
                ),
                helper.make_node(
                    "Gather",
                    ["present_0_decoder", "index_1"],
                    ["present.0.decoder.value"],
                    "Gather_1",
                    axis=0,
                ),
            ]
        )
    else:
        # Create nodes for Q/K/V paths
        q_nodes = [
            helper.make_node(
                "MatMul", ["layernorm_output_to_matmul", "q_weight"], ["q_matmul_output"], "q_path_matmul"
            ),
            helper.make_node("Add", ["q_bias", "q_matmul_output"], ["q_add_output"], "q_path_add"),
            helper.make_node("Mul", ["q_add_output", "q_scale"], ["q_mul_output"], "q_path_mul"),
            helper.make_node("Reshape", ["q_mul_output", "q_bsnh_reshape"], ["q_4d_bsnh"], "q_reshape_to_4d"),
            helper.make_node("Transpose", ["q_4d_bsnh"], ["q_4d_bnsh"], "q_transpose_to_bnsh", perm=[0, 2, 1, 3]),
            helper.make_node(
                "Reshape",
                ["q_4d_bnsh", "q_attn_heads_output"],
                ["q_output_(num_heads*batch_size,seq_len,head_size)"],
                "q_reshape_to_3d",
            ),
        ]
        k_nodes = [
            helper.make_node(
                "MatMul", ["layernorm_output_to_matmul", "k_weight"], ["k_matmul_output"], "k_path_matmul"
            ),
        ]
        if add_k:
            k_nodes.extend(
                [
                    helper.make_node("Add", ["k_bias", "k_matmul_output"], ["k_add_output"], "k_path_add"),
                    helper.make_node("Reshape", ["k_add_output", "bsnh_reshape"], ["k_4d_bsnh"], "k_reshape_to_4d"),
                ]
            )
        else:
            k_nodes.append(
                helper.make_node("Reshape", ["k_matmul_output", "kv_bsnh_reshape"], ["k_4d_bsnh"], "k_reshape_to_4d"),
            )
        k_nodes.extend(
            [
                helper.make_node(
                    "Transpose",
                    ["k_4d_bsnh"],
                    ["present.0.decoder.key"],
                    "k_transpose_to_bnsh",
                    perm=[0, 2, 1, 3],
                ),
                helper.make_node(
                    "Reshape",
                    ["present.0.decoder.key", "k_attn_heads_output"],
                    ["k_output_(num_heads*batch_size,seq_len,head_size)"],
                    "k_reshape_to_3d",
                ),
                helper.make_node(
                    "Transpose",
                    ["k_output_(num_heads*batch_size,seq_len,head_size)"],
                    ["k_output_(num_heads*batch_size,head_size,seq_len)"],
                    "k_transpose_last_two_dims",
                    perm=[0, 2, 1],
                ),
            ]
        )
        v_nodes = [
            helper.make_node(
                "MatMul", ["layernorm_output_to_matmul", "v_weight"], ["v_matmul_output"], "v_path_matmul"
            ),
            helper.make_node("Add", ["v_bias", "v_matmul_output"], ["v_add_output"], "v_path_add"),
            helper.make_node("Reshape", ["v_add_output", "kv_bsnh_reshape"], ["v_4d_bsnh"], "v_reshape_to_4d"),
            helper.make_node(
                "Transpose", ["v_4d_bsnh"], ["present.0.decoder.value"], "v_transpose_to_bnsh", perm=[0, 2, 1, 3]
            ),
            helper.make_node(
                "Reshape",
                ["present.0.decoder.value", "v_attn_heads_output"],
                ["v_output_(num_heads*batch_size,seq_len,head_size)"],
                "v_reshape_to_3d",
            ),
        ]
        nodes.extend(q_nodes)
        nodes.extend(k_nodes)
        nodes.extend(v_nodes)

        # Create nodes used with qkv concats, reshapes, and transposes
        nodes.extend(
            [
                helper.make_node("Shape", ["layernorm_output_to_matmul"], ["shape_output"], "shape"),
                helper.make_node("Gather", ["shape_output", "idx_0"], ["gather_0_output"], "gather_0", axis=0),
                helper.make_node(
                    "Mul", ["gather_0_output", "num_heads_int"], ["mul_attn_heads_output"], "mul_num_heads"
                ),
                helper.make_node(
                    "Unsqueeze",
                    ["mul_attn_heads_output", "unsqueeze_axes_input"],
                    ["unsqueeze_attn_heads_output"],
                    "unsqueeze_num_heads",
                ),
                helper.make_node(
                    "Concat",
                    ["unsqueeze_attn_heads_output", "neg_one", "head_size"],
                    ["q_attn_heads_output"],
                    "q_num_heads",
                    axis=0,
                ),
                helper.make_node(
                    "Concat",
                    ["unsqueeze_attn_heads_output", "neg_one", "head_size"],
                    ["k_attn_heads_output"],
                    "k_num_heads",
                    axis=0,
                ),
                helper.make_node(
                    "Concat",
                    ["unsqueeze_attn_heads_output", "neg_one", "head_size"],
                    ["v_attn_heads_output"],
                    "v_num_heads",
                    axis=0,
                ),
                helper.make_node(
                    "Constant",
                    inputs=[],
                    outputs=["q_bsnh_reshape"],
                    value=numpy_helper.from_array(
                        np.array([0, 0, num_heads, head_size], dtype="int64"), name="const_tensor"
                    ),
                ),
                helper.make_node(
                    "Constant",
                    inputs=[],
                    outputs=["kv_bsnh_reshape"],
                    value=numpy_helper.from_array(
                        np.array([0, -1, num_heads, head_size], dtype="int64"), name="const_tensor"
                    ),
                ),
            ]
        )

        # Create nodes used with mask
        nodes.extend(
            [
                helper.make_node(
                    "Shape", ["k_output_(num_heads*batch_size,seq_len,head_size)"], ["mask_shape_output"], "mask_shape"
                ),
                helper.make_node(
                    "Gather", ["mask_shape_output", "idx_1"], ["mask_gather_1_output"], "mask_gather_1", axis=0
                ),
                helper.make_node(
                    "Unsqueeze",
                    ["mask_gather_1_output", "unsqueeze_axes_input"],
                    ["mask_unsqueeze_1_output"],
                    "mask_unsqueeze_1",
                ),
                helper.make_node(
                    "Concat",
                    ["unsqueeze_0_output", "num_heads", "unsqueeze_1_output", "mask_unsqueeze_1_output"],
                    ["mask_concat_output"],
                    "mask_concat",
                    axis=0,
                ),
                helper.make_node(
                    "Mul", ["gather_0_output", "num_heads_int"], ["mul_mask_heads_output"], "mul_mask_heads"
                ),
                helper.make_node(
                    "Unsqueeze",
                    ["mul_mask_heads_output", "unsqueeze_axes_input"],
                    ["unsqueeze_mask_heads_output"],
                    "unsqueeze_mask_heads",
                ),
                helper.make_node(
                    "Concat",
                    ["unsqueeze_mask_heads_output", "unsqueeze_1_output", "mask_unsqueeze_1_output"],
                    ["concat_input_for_reshape_after_add"],
                    "concat_for_reshape_after_add",
                    axis=0,
                ),
            ]
        )

        # Create nodes used with Q x K' + mask and softmax(Q x K' + mask) x V
        nodes.extend(
            [
                helper.make_node("Gather", ["shape_output", "idx_1"], ["gather_1_output"], "gather_1", axis=0),
                helper.make_node(
                    "Unsqueeze", ["gather_0_output", "unsqueeze_axes_input"], ["unsqueeze_0_output"], "unsqueeze_0"
                ),
                helper.make_node(
                    "Unsqueeze", ["gather_1_output", "unsqueeze_axes_input"], ["unsqueeze_1_output"], "unsqueeze_1"
                ),
                helper.make_node(
                    "Concat",
                    ["unsqueeze_0_output", "num_heads", "unsqueeze_1_output", "head_size"],
                    ["bnsh_format"],
                    axis=0,
                ),
                helper.make_node(
                    "Concat", ["unsqueeze_0_output", "unsqueeze_1_output", "hidden_size"], ["bsd_format"], axis=0
                ),
            ]
        )

        # Create nodes for computing softmax(Q x K' + mask) x V
        nodes.extend(
            [
                helper.make_node(
                    "MatMul",
                    [
                        "q_output_(num_heads*batch_size,seq_len,head_size)",
                        "k_output_(num_heads*batch_size,head_size,seq_len)",
                    ],
                    ["qk_output_(num_heads*batch_size,seq_len,seq_len)"],
                    "matmul_qk",
                ),
                helper.make_node(
                    "Reshape",
                    ["qk_output_(num_heads*batch_size,seq_len,seq_len)", "mask_concat_output"],
                    ["qk_output_(batch_size,num_heads,seq_len,seq_len)"],
                    "reshape_qk_to_bnsh",
                ),
                helper.make_node(
                    "Add",
                    ["qk_output_(batch_size,num_heads,seq_len,seq_len)", "attention_add_qk"],
                    ["add_qk_output_(batch_size,num_heads_seq_len,seq_len)"],
                    "add_qk",
                ),
                helper.make_node(
                    "Reshape",
                    ["add_qk_output_(batch_size,num_heads_seq_len,seq_len)", "concat_input_for_reshape_after_add"],
                    ["add_qk_output_(num_heads*batch_size,seq_len,seq_len)"],
                    "reshape_add_qk_before_softmax",
                ),
                helper.make_node(
                    "Softmax",
                    ["add_qk_output_(num_heads*batch_size,seq_len,seq_len)"],
                    ["softmax_output"],
                    "softmax_qk",
                    axis=2,
                ),
                helper.make_node(
                    "MatMul",
                    ["softmax_output", "v_output_(num_heads*batch_size,seq_len,head_size)"],
                    ["qkv_output_(num_heads*batch_size,seq_len,head_size)"],
                    "matmul_qkv",
                ),
                helper.make_node(
                    "Reshape",
                    ["qkv_output_(num_heads*batch_size,seq_len,head_size)", "bnsh_format"],
                    ["qkv_bnsh"],
                    "reshape_qkv_to_bnsh",
                ),
                helper.make_node("Transpose", ["qkv_bnsh"], ["qkv_bsnh"], "transpose_bnsh_to_bsnh", perm=[0, 2, 1, 3]),
                helper.make_node("Reshape", ["qkv_bsnh", "bsd_format"], ["attn_output"], "qkv_bsd"),
            ]
        )

    # Create nodes that make attention mask
    nodes.extend(
        [
            # "attention_mask" is (decoder_seq_len, decoder_seq_len) but is assumed to be (1, 1) for this test.
            # There are other nodes that automatically set the attention mask size correctly but those nodes do not
            # impact the attention fusion. Hence, this assumption is made in order to simplify the inputs for the
            # following nodes.
            helper.make_node(
                "Where",
                ["all_ones", "where_filter_constant", "dummy_input_fp32"],
                ["where_output"],
                "mask_filter_where",
            ),
            helper.make_node(
                "Unsqueeze",
                ["where_output", "dummy_input_int64"],
                ["unsqueeze_mask_output_1"],
                "unsqueeze_attn_mask_1",
            ),
            helper.make_node(
                "Unsqueeze",
                ["unsqueeze_mask_output_1", "dummy_input_int64"],
                ["unsqueeze_mask_output_2"],
                "unsqueeze_attn_mask_2",
            ),
            helper.make_node(
                "Expand",
                inputs=["unsqueeze_mask_output_2", "dummy_input_int64"],
                outputs=["attention_add_qk"],
                name="expand_mask_from_(b,1,m,m)_to_(b,n,m,m)",
            ),
        ]
    )

    # Create final nodes to conclude attention
    nodes.append(
        helper.make_node(
            "MatMul",
            ["attn_output", "matmul_after_attn_initializer"],
            ["matmul_after_attn_output"],
            "matmul_after_attn",
        ),
    )
    if not fused:
        next_sln_inputs = [
            "layernorm_output_to_skiplayernorm",
            "add_after_attn_output",
            "layernorm_weight",
            "layernorm_bias",
        ]
        nodes.extend(
            [
                helper.make_node(
                    "Add",
                    ["add_after_attn_initializer", "matmul_after_attn_output"],
                    ["add_after_attn_output"],
                    "add_after_attn",
                ),
                helper.make_node(
                    "SkipLayerNormalization",
                    next_sln_inputs,
                    ["output_0", "", "", "output_1"],
                    "next_skiplayernorm",
                    domain="com.microsoft",
                    epsilon=epsilon,
                ),
            ]
        )
    else:
        next_sln_inputs = [
            "matmul_after_attn_output",
            "layernorm_output_to_skiplayernorm",
            "layernorm_weight",
            "layernorm_bias",
            "add_after_attn_initializer",
        ]
        nodes.append(
            helper.make_node(
                "SkipLayerNormalization",
                next_sln_inputs,
                ["output_0", "", "", "output_1"],
                "SkipLayerNorm_AddBias_0",
                domain="com.microsoft",
                epsilon=epsilon,
            )
        )

    # Create initializers
    q_weight, q_weight_data = get_tensor_and_weight("q_weight", [hidden_size, hidden_size])
    q_bias, q_bias_data = get_tensor_and_weight("q_bias", [hidden_size])
    k_weight, k_weight_data = get_tensor_and_weight("k_weight", [hidden_size, hidden_size])
    k_bias, k_bias_data = get_tensor_and_weight("k_bias", [hidden_size], zeros=(not add_k))
    v_weight, v_weight_data = get_tensor_and_weight("v_weight", [hidden_size, hidden_size])
    v_bias, v_bias_data = get_tensor_and_weight("v_bias", [hidden_size])
    qkv_weight = helper.make_tensor(
        "Attention_0_qkv_weight",
        TensorProto.FLOAT,
        [hidden_size, 3 * hidden_size],
        q_weight_data + k_weight_data + v_weight_data,
    )
    qkv_bias = helper.make_tensor(
        "Attention_0_qkv_bias", TensorProto.FLOAT, [3 * hidden_size], q_bias_data + k_bias_data + v_bias_data
    )
    initializers = [
        float_tensor("layernorm_weight", [hidden_size]),
        float_tensor("layernorm_bias", [hidden_size]),
        float_tensor("matmul_after_attn_initializer", [hidden_size, hidden_size]),
        float_tensor("add_after_attn_initializer", [hidden_size]),
    ]
    # Add initializers for attention mask
    initializers.extend(
        [
            numpy_helper.from_array(np.array([[1]], dtype=bool), name="all_ones"),
            numpy_helper.from_array(np.array([1], dtype="float32"), name="where_filter_constant"),
        ]
    )

    if fused:
        initializers.extend(
            [
                qkv_weight,
                qkv_bias,
                numpy_helper.from_array(np.array(0, dtype="int64"), name="index_0"),
                numpy_helper.from_array(np.array(1, dtype="int64"), name="index_1"),
            ]
        )
    else:
        initializers.extend(
            [
                numpy_helper.from_array(np.array(num_heads, dtype="int64"), name="num_heads_int"),
                numpy_helper.from_array(np.array([num_heads], dtype="int64"), name="num_heads"),
                numpy_helper.from_array(np.array([head_size], dtype="int64"), name="head_size"),
                numpy_helper.from_array(np.array([hidden_size], dtype="int64"), name="hidden_size"),
                numpy_helper.from_array(np.array(1 / np.sqrt(head_size), dtype="float32"), name="q_scale"),
                numpy_helper.from_array(np.array(0, dtype="int64"), name="idx_0"),
                numpy_helper.from_array(np.array(1, dtype="int64"), name="idx_1"),
                numpy_helper.from_array(np.array([-1], dtype="int64"), name="neg_one"),
                numpy_helper.from_array(np.array([0], dtype="int64"), name="unsqueeze_axes_input"),
            ]
        )

        if add_k:
            initializers.extend([q_weight, q_bias, k_weight, k_bias, v_weight, v_bias])
        else:
            initializers.extend([q_weight, q_bias, k_weight, v_weight, v_bias])

    # Construct graph
    graph = helper.make_graph(
        nodes, "whisper_decoder_attention_graph", inputs, outputs, initializers, doc_string="whisper"
    )
    opsetid = helper.make_opsetid("ai.onnx", min(onnx.defs.onnx_opset_version(), 16))
    return helper.make_model(graph, opset_imports=(opsetid,))


def create_whisper_decoder_multihead_attention(
    hidden_size=768, num_heads=12, epsilon=0.000009999999747378752, add_k=False, fused=False
):
    # Get head size and ensure head size is an integer
    assert hidden_size % num_heads == 0
    head_size = hidden_size // num_heads

    # Construct input and output nodes
    inputs = [
        helper.make_tensor_value_info("input_0", TensorProto.FLOAT, ["batch_size", 1500, hidden_size]),
        helper.make_tensor_value_info("encoder_hidden_states", TensorProto.FLOAT, ["batch_size", 1500, hidden_size]),
    ]
    outputs = [
        helper.make_tensor_value_info("output_0", TensorProto.FLOAT, ["batch_size", 1500, hidden_size]),
        helper.make_tensor_value_info("output_1", TensorProto.FLOAT, ["batch_size", "sequence_length", hidden_size]),
        helper.make_tensor_value_info(
            "present.0.encoder.key", TensorProto.FLOAT, ["batch_size", num_heads, 1500, head_size]
        ),
        helper.make_tensor_value_info(
            "present.0.encoder.value", TensorProto.FLOAT, ["batch_size", num_heads, 1500, head_size]
        ),
    ]

    # Create SkipLayerNorm (since there's no Add + LayerNorm variant for this attention subgraph)
    nodes = [
        helper.make_node(
            "SkipLayerNormalization",
            ["input_0", "input_0", "layernorm_weight", "layernorm_bias"],
            ["layernorm_output_to_matmul", "", "", "layernorm_output_to_skiplayernorm"],
            "skiplayernorm",
            domain="com.microsoft",
            epsilon=epsilon,
        )
    ]

    if fused:
        nodes.extend(
            [
                helper.make_node(
                    "MatMul", ["layernorm_output_to_matmul", "q_weight"], ["q_matmul_output"], "q_path_matmul"
                ),
                helper.make_node("MatMul", ["encoder_hidden_states", "k_weight"], ["k_matmul_output"], "k_path_matmul"),
                helper.make_node("MatMul", ["encoder_hidden_states", "v_weight"], ["v_matmul_output"], "v_path_matmul"),
                helper.make_node(
                    "MultiHeadAttention",
                    ["q_matmul_output", "k_matmul_output", "v_matmul_output", "Attention_0_qkv_bias"],
                    ["attn_output", "present.0.encoder.key", "present.0.encoder.value"],
                    "Attention_0",
                    domain="com.microsoft",
                    num_heads=num_heads,
                ),
            ]
        )
    else:
        # Create nodes for Q/K/V paths
        q_nodes = [
            helper.make_node(
                "MatMul",
                ["layernorm_output_to_matmul", "q_weight"],
                ["q_matmul_output"],
                "q_path_matmul",
            ),
            helper.make_node("Add", ["q_bias", "q_matmul_output"], ["q_add_output"], "q_path_add"),
            helper.make_node("Mul", ["q_add_output", "q_scale"], ["q_mul_output"], "q_path_mul"),
            helper.make_node("Reshape", ["q_mul_output", "q_bsnh_reshape"], ["q_4d_bsnh"], "q_reshape_to_4d"),
            helper.make_node("Transpose", ["q_4d_bsnh"], ["q_4d_bnsh"], "q_transpose_to_bnsh", perm=[0, 2, 1, 3]),
            helper.make_node(
                "Reshape",
                ["q_4d_bnsh", "q_attn_heads_output"],
                ["q_output_(num_heads*batch_size,seq_len,head_size)"],
                "q_reshape_to_3d",
            ),
        ]
        k_nodes = [
            helper.make_node("MatMul", ["encoder_hidden_states", "k_weight"], ["k_matmul_output"], "k_path_matmul"),
        ]
        if add_k:
            k_nodes.extend(
                [
                    helper.make_node("Add", ["k_bias", "k_matmul_output"], ["k_add_output"], "k_path_add"),
                    helper.make_node("Reshape", ["k_add_output", "bsnh_reshape"], ["k_4d_bsnh"], "k_reshape_to_4d"),
                ]
            )
        else:
            k_nodes.append(
                helper.make_node("Reshape", ["k_matmul_output", "kv_bsnh_reshape"], ["k_4d_bsnh"], "k_reshape_to_4d"),
            )
        k_nodes.extend(
            [
                helper.make_node(
                    "Transpose", ["k_4d_bsnh"], ["present.0.encoder.key"], "k_transpose_to_bnsh", perm=[0, 2, 1, 3]
                ),
                helper.make_node(
                    "Reshape",
                    ["present.0.encoder.key", "k_attn_heads_output"],
                    ["k_output_(num_heads*batch_size,seq_len,head_size)"],
                    "k_reshape_to_3d",
                ),
                helper.make_node(
                    "Transpose",
                    ["k_output_(num_heads*batch_size,seq_len,head_size)"],
                    ["k_output_(num_heads*batch_size,head_size,seq_len)"],
                    "k_transpose_last_two_dims",
                    perm=[0, 2, 1],
                ),
            ]
        )
        v_nodes = [
            helper.make_node("MatMul", ["encoder_hidden_states", "v_weight"], ["v_matmul_output"], "v_path_matmul"),
            helper.make_node("Add", ["v_bias", "v_matmul_output"], ["v_add_output"], "v_path_add"),
            helper.make_node("Reshape", ["v_add_output", "kv_bsnh_reshape"], ["v_4d_bsnh"], "v_reshape_to_4d"),
            helper.make_node(
                "Transpose", ["v_4d_bsnh"], ["present.0.encoder.value"], "v_transpose_to_bnsh", perm=[0, 2, 1, 3]
            ),
            helper.make_node(
                "Reshape",
                ["present.0.encoder.value", "v_attn_heads_output"],
                ["v_output_(num_heads*batch_size,seq_len,head_size)"],
                "v_reshape_to_3d",
            ),
        ]
        nodes.extend(q_nodes)
        nodes.extend(k_nodes)
        nodes.extend(v_nodes)

        # Create nodes used with qkv concats, reshapes, and transposes
        nodes.extend(
            [
                helper.make_node("Shape", ["layernorm_output_to_matmul"], ["shape_output"], "shape"),
                helper.make_node("Gather", ["shape_output", "idx_0"], ["gather_0_output"], "gather_0", axis=0),
                helper.make_node(
                    "Mul", ["gather_0_output", "num_heads_int"], ["mul_attn_heads_output"], "mul_num_heads"
                ),
                helper.make_node(
                    "Unsqueeze",
                    ["mul_attn_heads_output", "unsqueeze_axes_input"],
                    ["unsqueeze_attn_heads_output"],
                    "unsqueeze_num_heads",
                ),
                helper.make_node(
                    "Concat",
                    ["unsqueeze_attn_heads_output", "neg_one", "head_size"],
                    ["q_attn_heads_output"],
                    "q_num_heads",
                    axis=0,
                ),
                helper.make_node(
                    "Concat",
                    ["unsqueeze_attn_heads_output", "neg_one", "head_size"],
                    ["k_attn_heads_output"],
                    "k_num_heads",
                    axis=0,
                ),
                helper.make_node(
                    "Concat",
                    ["unsqueeze_attn_heads_output", "neg_one", "head_size"],
                    ["v_attn_heads_output"],
                    "v_num_heads",
                    axis=0,
                ),
                helper.make_node(
                    "Constant",
                    inputs=[],
                    outputs=["q_bsnh_reshape"],
                    value=numpy_helper.from_array(
                        np.array([0, 0, num_heads, head_size], dtype="int64"), name="const_tensor"
                    ),
                ),
                helper.make_node(
                    "Constant",
                    inputs=[],
                    outputs=["kv_bsnh_reshape"],
                    value=numpy_helper.from_array(
                        np.array([0, -1, num_heads, head_size], dtype="int64"), name="const_tensor"
                    ),
                ),
            ]
        )

        # Create nodes used with Q x K' and softmax(Q x K') x V
        nodes.extend(
            [
                helper.make_node("Gather", ["shape_output", "idx_1"], ["gather_1_output"], "gather_1", axis=0),
                helper.make_node(
                    "Unsqueeze",
                    ["gather_0_output", "unsqueeze_axes_input"],
                    ["unsqueeze_0_output"],
                    "unsqueeze_0",
                ),
                helper.make_node(
                    "Unsqueeze",
                    ["gather_1_output", "unsqueeze_axes_input"],
                    ["unsqueeze_1_output"],
                    "unsqueeze_1",
                ),
                helper.make_node(
                    "Concat",
                    ["unsqueeze_0_output", "num_heads", "unsqueeze_1_output", "head_size"],
                    ["bnsh_format"],
                    axis=0,
                ),
                helper.make_node(
                    "Concat",
                    ["unsqueeze_0_output", "unsqueeze_1_output", "hidden_size"],
                    ["bsd_format"],
                    axis=0,
                ),
            ]
        )

        # Create nodes for computing softmax(Q x K') x V
        nodes.extend(
            [
                helper.make_node(
                    "MatMul",
                    [
                        "q_output_(num_heads*batch_size,seq_len,head_size)",
                        "k_output_(num_heads*batch_size,head_size,seq_len)",
                    ],
                    ["qk_output_(num_heads*batch_size,seq_len,seq_len)"],
                    "matmul_qk",
                ),
                helper.make_node(
                    "Softmax",
                    ["qk_output_(num_heads*batch_size,seq_len,seq_len)"],
                    ["softmax_output"],
                    "softmax_qk",
                    axis=2,
                ),
                helper.make_node(
                    "MatMul",
                    ["softmax_output", "v_output_(num_heads*batch_size,seq_len,head_size)"],
                    ["qkv_output_(num_heads*batch_size,seq_len,head_size)"],
                    "matmul_qkv",
                ),
                helper.make_node(
                    "Reshape",
                    ["qkv_output_(num_heads*batch_size,seq_len,head_size)", "bnsh_format"],
                    ["qkv_bnsh"],
                    "reshape_qkv_to_bnsh",
                ),
                helper.make_node("Transpose", ["qkv_bnsh"], ["qkv_bsnh"], "transpose_bnsh_to_bsnh", perm=[0, 2, 1, 3]),
                helper.make_node("Reshape", ["qkv_bsnh", "bsd_format"], ["attn_output"], "qkv_bsd"),
            ]
        )

    # Create final nodes to conclude attention
    nodes.append(
        helper.make_node(
            "MatMul",
            ["attn_output", "matmul_after_attn_initializer"],
            ["matmul_after_attn_output"],
            "matmul_after_attn",
        ),
    )
    if not fused:
        next_sln_inputs = [
            "layernorm_output_to_skiplayernorm",
            "add_after_attn_output",
            "layernorm_weight",
            "layernorm_bias",
        ]
        nodes.extend(
            [
                helper.make_node(
                    "Add",
                    ["add_after_attn_initializer", "matmul_after_attn_output"],
                    ["add_after_attn_output"],
                    "add_after_attn",
                ),
                helper.make_node(
                    "SkipLayerNormalization",
                    next_sln_inputs,
                    ["output_0", "", "", "output_1"],
                    "next_skiplayernorm",
                    domain="com.microsoft",
                    epsilon=epsilon,
                ),
            ]
        )
    else:
        next_sln_inputs = [
            "matmul_after_attn_output",
            "layernorm_output_to_skiplayernorm",
            "layernorm_weight",
            "layernorm_bias",
            "add_after_attn_initializer",
        ]
        nodes.append(
            helper.make_node(
                "SkipLayerNormalization",
                next_sln_inputs,
                ["output_0", "", "", "output_1"],
                "SkipLayerNorm_AddBias_0",
                domain="com.microsoft",
                epsilon=epsilon,
            )
        )

    # Create initializers
    q_weight, q_weight_data = get_tensor_and_weight("q_weight", [hidden_size, hidden_size])
    q_bias, q_bias_data = get_tensor_and_weight("q_bias", [hidden_size])
    k_weight, k_weight_data = get_tensor_and_weight("k_weight", [hidden_size, hidden_size])
    k_bias, k_bias_data = get_tensor_and_weight("k_bias", [hidden_size], zeros=(not add_k))
    v_weight, v_weight_data = get_tensor_and_weight("v_weight", [hidden_size, hidden_size])
    v_bias, v_bias_data = get_tensor_and_weight("v_bias", [hidden_size])
    qkv_bias = helper.make_tensor(
        "Attention_0_qkv_bias", TensorProto.FLOAT, [3 * hidden_size], q_bias_data + k_bias_data + v_bias_data
    )
    initializers = [
        float_tensor("layernorm_weight", [hidden_size]),
        float_tensor("layernorm_bias", [hidden_size]),
        float_tensor("matmul_after_attn_initializer", [hidden_size, hidden_size]),
        float_tensor("add_after_attn_initializer", [hidden_size]),
    ]

    # Add Q/K/V weight tensors as initializers
    initializers.extend([q_weight, k_weight, v_weight])

    if fused:
        initializers.append(qkv_bias)
    else:
        if add_k:
            initializers.extend([q_bias, k_bias, v_bias])
        else:
            initializers.extend([q_bias, v_bias])

        initializers.extend(
            [
                numpy_helper.from_array(np.array(num_heads, dtype="int64"), name="num_heads_int"),
                numpy_helper.from_array(np.array([num_heads], dtype="int64"), name="num_heads"),
                numpy_helper.from_array(np.array([head_size], dtype="int64"), name="head_size"),
                numpy_helper.from_array(np.array([hidden_size], dtype="int64"), name="hidden_size"),
                numpy_helper.from_array(np.array(1 / np.sqrt(head_size), dtype="float32"), name="q_scale"),
                numpy_helper.from_array(np.array(0, dtype="int64"), name="idx_0"),
                numpy_helper.from_array(np.array(1, dtype="int64"), name="idx_1"),
                numpy_helper.from_array(np.array([-1], dtype="int64"), name="neg_one"),
                numpy_helper.from_array(np.array([0], dtype="int64"), name="unsqueeze_axes_input"),
            ]
        )

    # Construct graph
    graph = helper.make_graph(nodes, "whisper_decoder_mha_graph", inputs, outputs, initializers, doc_string="whisper")
    opsetid = helper.make_opsetid("ai.onnx", min(onnx.defs.onnx_opset_version(), 16))
    return helper.make_model(graph, opset_imports=(opsetid,))


def create_whisper_decoder_with_past_multihead_self_attention(
    hidden_size=768,
    num_heads=12,
    epsilon=0.000009999999747378752,
    add_before_layernorm=False,
    add_k=False,
    fused=False,
):
    # Get head size and ensure head size is an integer
    assert hidden_size % num_heads == 0
    head_size = hidden_size // num_heads

    # Construct input and output nodes
    inputs = [
        helper.make_tensor_value_info("input_0", TensorProto.FLOAT, ["batch_size", 1500, hidden_size]),
        helper.make_tensor_value_info(
            "past_key_values.0.decoder.key", TensorProto.FLOAT, ["batch_size", num_heads, "past_seq_len", head_size]
        ),
        helper.make_tensor_value_info(
            "past_key_values.0.decoder.value", TensorProto.FLOAT, ["batch_size", num_heads, "past_seq_len", head_size]
        ),
    ]
    outputs = [
        helper.make_tensor_value_info("output_0", TensorProto.FLOAT, ["batch_size", 1500, hidden_size]),
        helper.make_tensor_value_info("output_1", TensorProto.FLOAT, ["batch_size", "sequence_length", hidden_size]),
        helper.make_tensor_value_info(
            "present.0.decoder.key", TensorProto.FLOAT, ["batch_size", num_heads, "past_seq_len + 1", head_size]
        ),
        helper.make_tensor_value_info(
            "present.0.decoder.value", TensorProto.FLOAT, ["batch_size", num_heads, "past_seq_len + 1", head_size]
        ),
    ]
    nodes = []

    # Create layernorm (Add + LayerNorm or SkipLayerNorm)
    if add_before_layernorm:
        nodes.extend(
            [
                helper.make_node(
                    "Add", ["input_0", "input_0"], ["layernorm_output_to_skiplayernorm"], "add_before_layernorm"
                ),
                helper.make_node(
                    "LayerNormalization",
                    ["layernorm_output_to_skiplayernorm", "layernorm_weight", "layernorm_bias"],
                    ["layernorm_output_to_matmul"],
                    "layernorm",
                    epsilon=epsilon,
                ),
            ]
        )
    else:
        nodes.append(
            helper.make_node(
                "SkipLayerNormalization",
                ["input_0", "input_0", "layernorm_weight", "layernorm_bias"],
                ["layernorm_output_to_matmul", "", "", "layernorm_output_to_skiplayernorm"],
                "skiplayernorm",
                domain="com.microsoft",
                epsilon=epsilon,
            )
        )

    if fused:
        nodes.extend(
            [
                helper.make_node(
                    "MatMul",
                    ["layernorm_output_to_matmul", "MatMul_0_qkv_weight"],
                    ["MatMul_0_qkv_out"],
                    "MatMul_0",
                ),
                helper.make_node(
                    "Slice",
                    ["MatMul_0_qkv_out", "MatMul_0_q_start_index", "MatMul_0_k_start_index", "MatMul_0_qkv_last_axis"],
                    ["MatMul_0_q_out"],
                    "Slice_0",
                ),
                helper.make_node(
                    "Slice",
                    ["MatMul_0_qkv_out", "MatMul_0_k_start_index", "MatMul_0_v_start_index", "MatMul_0_qkv_last_axis"],
                    ["MatMul_0_k_out"],
                    "Slice_1",
                ),
                helper.make_node(
                    "Slice",
                    [
                        "MatMul_0_qkv_out",
                        "MatMul_0_v_start_index",
                        "MatMul_0_end_of_qkv_index",
                        "MatMul_0_qkv_last_axis",
                    ],
                    ["MatMul_0_v_out"],
                    "Slice_2",
                ),
                helper.make_node(
                    "MultiHeadAttention",
                    [
                        "MatMul_0_q_out",
                        "MatMul_0_k_out",
                        "MatMul_0_v_out",
                        "Attention_0_qkv_bias",
                        "",
                        "",
                        "past_key_values.0.decoder.key",
                        "past_key_values.0.decoder.value",
                    ],
                    ["attn_output", "present.0.decoder.key", "present.0.decoder.value"],
                    "Attention_0",
                    domain="com.microsoft",
                    num_heads=num_heads,
                ),
            ]
        )
    else:
        # Create nodes for Q/K/V paths
        q_nodes = [
            helper.make_node(
                "MatMul", ["layernorm_output_to_matmul", "q_weight"], ["q_matmul_output"], "q_path_matmul"
            ),
            helper.make_node("Add", ["q_bias", "q_matmul_output"], ["q_add_output"], "q_path_add"),
            helper.make_node("Mul", ["q_add_output", "q_scale"], ["q_mul_output"], "q_path_mul"),
            helper.make_node("Reshape", ["q_mul_output", "q_bsnh_reshape"], ["q_4d_bsnh"], "q_reshape_to_4d"),
            helper.make_node("Transpose", ["q_4d_bsnh"], ["q_4d_bnsh"], "q_transpose_to_bnsh", perm=[0, 2, 1, 3]),
            helper.make_node(
                "Reshape",
                ["q_4d_bnsh", "q_attn_heads_output"],
                ["q_output_(num_heads*batch_size,seq_len,head_size)"],
                "q_reshape_to_3d",
            ),
        ]
        k_nodes = [
            helper.make_node(
                "MatMul",
                ["layernorm_output_to_matmul", "k_weight"],
                ["k_matmul_output"],
                "k_path_matmul",
            ),
        ]
        if add_k:
            k_nodes.extend(
                [
                    helper.make_node("Add", ["k_bias", "k_matmul_output"], ["k_add_output"], "k_path_add"),
                    helper.make_node("Reshape", ["k_add_output", "bsnh_reshape"], ["k_4d_bsnh"], "k_reshape_to_4d"),
                ]
            )
        else:
            k_nodes.append(
                helper.make_node("Reshape", ["k_matmul_output", "kv_bsnh_reshape"], ["k_4d_bsnh"], "k_reshape_to_4d"),
            )
        k_nodes.extend(
            [
                helper.make_node("Transpose", ["k_4d_bsnh"], ["k_4d_bnsh"], "k_transpose_to_bnsh", perm=[0, 2, 1, 3]),
                helper.make_node(
                    "Concat",
                    ["past_key_values.0.decoder.key", "k_4d_bnsh"],
                    ["present.0.decoder.key"],
                    "concat_past_k_and_curr_k",
                    axis=2,
                ),
                helper.make_node(
                    "Reshape",
                    ["present.0.decoder.key", "k_attn_heads_output"],
                    ["k_output_(num_heads*batch_size,seq_len,head_size)"],
                    "k_reshape_to_3d",
                ),
                helper.make_node(
                    "Transpose",
                    ["k_output_(num_heads*batch_size,seq_len,head_size)"],
                    ["k_output_(num_heads*batch_size,head_size,seq_len)"],
                    "k_transpose_last_two_dims",
                    perm=[0, 2, 1],
                ),
            ]
        )
        v_nodes = [
            helper.make_node(
                "MatMul",
                ["layernorm_output_to_matmul", "v_weight"],
                ["v_matmul_output"],
                "v_path_matmul",
            ),
            helper.make_node("Add", ["v_bias", "v_matmul_output"], ["v_add_output"], "v_path_add"),
            helper.make_node("Reshape", ["v_add_output", "kv_bsnh_reshape"], ["v_4d_bsnh"], "v_reshape_to_4d"),
            helper.make_node("Transpose", ["v_4d_bsnh"], ["v_4d_bnsh"], "v_transpose_to_bnsh", perm=[0, 2, 1, 3]),
            helper.make_node(
                "Concat",
                ["past_key_values.0.decoder.value", "v_4d_bnsh"],
                ["present.0.decoder.value"],
                "concat_past_v_and_curr_v",
                axis=2,
            ),
            helper.make_node(
                "Reshape",
                ["present.0.decoder.value", "v_attn_heads_output"],
                ["v_output_(num_heads*batch_size,seq_len,head_size)"],
                "v_reshape_to_3d",
            ),
        ]
        nodes.extend(q_nodes)
        nodes.extend(k_nodes)
        nodes.extend(v_nodes)

        # Create nodes used with qkv concats, reshapes, and transposes
        nodes.extend(
            [
                helper.make_node("Shape", ["layernorm_output_to_matmul"], ["shape_output"], "shape"),
                helper.make_node("Gather", ["shape_output", "idx_0"], ["gather_0_output"], "gather_0", axis=0),
                helper.make_node(
                    "Mul",
                    ["gather_0_output", "num_heads_int"],
                    ["mul_attn_heads_output"],
                    "mul_num_heads",
                ),
                helper.make_node(
                    "Unsqueeze",
                    ["mul_attn_heads_output", "unsqueeze_axes_input"],
                    ["unsqueeze_attn_heads_output"],
                    "unsqueeze_num_heads",
                ),
                helper.make_node(
                    "Concat",
                    ["unsqueeze_attn_heads_output", "neg_one", "head_size"],
                    ["q_attn_heads_output"],
                    "q_num_heads",
                    axis=0,
                ),
                helper.make_node(
                    "Concat",
                    ["unsqueeze_attn_heads_output", "neg_one", "head_size"],
                    ["k_attn_heads_output"],
                    "k_num_heads",
                    axis=0,
                ),
                helper.make_node(
                    "Concat",
                    ["unsqueeze_attn_heads_output", "neg_one", "head_size"],
                    ["v_attn_heads_output"],
                    "v_num_heads",
                    axis=0,
                ),
                helper.make_node(
                    "Constant",
                    inputs=[],
                    outputs=["q_bsnh_reshape"],
                    value=numpy_helper.from_array(
                        np.array([0, 0, num_heads, head_size], dtype="int64"), name="const_tensor"
                    ),
                ),
                helper.make_node(
                    "Constant",
                    inputs=[],
                    outputs=["kv_bsnh_reshape"],
                    value=numpy_helper.from_array(
                        np.array([0, -1, num_heads, head_size], dtype="int64"), name="const_tensor"
                    ),
                ),
            ]
        )

        # Create nodes used with Q x K' and softmax(Q x K') x V
        nodes.extend(
            [
                helper.make_node("Gather", ["shape_output", "idx_1"], ["gather_1_output"], "gather_1", axis=0),
                helper.make_node(
                    "Unsqueeze",
                    ["gather_0_output", "unsqueeze_axes_input"],
                    ["unsqueeze_0_output"],
                    "unsqueeze_0",
                ),
                helper.make_node(
                    "Unsqueeze",
                    ["gather_1_output", "unsqueeze_axes_input"],
                    ["unsqueeze_1_output"],
                    "unsqueeze_1",
                ),
                helper.make_node(
                    "Concat",
                    ["unsqueeze_0_output", "num_heads", "unsqueeze_1_output", "head_size"],
                    ["bnsh_format"],
                    axis=0,
                ),
                helper.make_node(
                    "Concat",
                    ["unsqueeze_0_output", "unsqueeze_1_output", "hidden_size"],
                    ["bsd_format"],
                    axis=0,
                ),
            ]
        )

        # Create nodes for computing softmax(Q x K') x V
        nodes.extend(
            [
                helper.make_node(
                    "MatMul",
                    [
                        "q_output_(num_heads*batch_size,seq_len,head_size)",
                        "k_output_(num_heads*batch_size,head_size,seq_len)",
                    ],
                    ["qk_output_(num_heads*batch_size,seq_len,seq_len)"],
                    "matmul_qk",
                ),
                helper.make_node(
                    "Softmax",
                    ["qk_output_(num_heads*batch_size,seq_len,seq_len)"],
                    ["softmax_output"],
                    "softmax_qk",
                    axis=2,
                ),
                helper.make_node(
                    "MatMul",
                    ["softmax_output", "v_output_(num_heads*batch_size,seq_len,head_size)"],
                    ["qkv_output_(num_heads*batch_size,seq_len,head_size)"],
                    "matmul_qkv",
                ),
                helper.make_node(
                    "Reshape",
                    ["qkv_output_(num_heads*batch_size,seq_len,head_size)", "bnsh_format"],
                    ["qkv_bnsh"],
                    "reshape_qkv_to_bnsh",
                ),
                helper.make_node("Transpose", ["qkv_bnsh"], ["qkv_bsnh"], "transpose_bnsh_to_bsnh", perm=[0, 2, 1, 3]),
                helper.make_node("Reshape", ["qkv_bsnh", "bsd_format"], ["attn_output"], "qkv_bsd"),
            ]
        )

    # Create final nodes to conclude attention
    nodes.append(
        helper.make_node(
            "MatMul",
            ["attn_output", "matmul_after_attn_initializer"],
            ["matmul_after_attn_output"],
            "matmul_after_attn",
        ),
    )
    if not fused:
        next_sln_inputs = [
            "layernorm_output_to_skiplayernorm",
            "add_after_attn_output",
            "layernorm_weight",
            "layernorm_bias",
        ]
        nodes.extend(
            [
                helper.make_node(
                    "Add",
                    ["add_after_attn_initializer", "matmul_after_attn_output"],
                    ["add_after_attn_output"],
                    "add_after_attn",
                ),
                helper.make_node(
                    "SkipLayerNormalization",
                    next_sln_inputs,
                    ["output_0", "", "", "output_1"],
                    "next_skiplayernorm",
                    domain="com.microsoft",
                    epsilon=epsilon,
                ),
            ]
        )
    else:
        next_sln_inputs = [
            "matmul_after_attn_output",
            "layernorm_output_to_skiplayernorm",
            "layernorm_weight",
            "layernorm_bias",
            "add_after_attn_initializer",
        ]
        nodes.append(
            helper.make_node(
                "SkipLayerNormalization",
                next_sln_inputs,
                ["output_0", "", "", "output_1"],
                "SkipLayerNorm_AddBias_0",
                domain="com.microsoft",
                epsilon=epsilon,
            )
        )

    # Create initializers
    q_weight, q_weight_data = get_tensor_and_weight("q_weight", [hidden_size, hidden_size])
    q_bias, q_bias_data = get_tensor_and_weight("q_bias", [hidden_size])
    k_weight, k_weight_data = get_tensor_and_weight("k_weight", [hidden_size, hidden_size])
    k_bias, k_bias_data = get_tensor_and_weight("k_bias", [hidden_size], zeros=(not add_k))
    v_weight, v_weight_data = get_tensor_and_weight("v_weight", [hidden_size, hidden_size])
    v_bias, v_bias_data = get_tensor_and_weight("v_bias", [hidden_size])
    qkv_weight = helper.make_tensor(
        "MatMul_0_qkv_weight",
        TensorProto.FLOAT,
        [hidden_size, 3 * hidden_size],
        q_weight_data + k_weight_data + v_weight_data,
    )
    qkv_bias = helper.make_tensor(
        "Attention_0_qkv_bias",
        TensorProto.FLOAT,
        [3 * hidden_size],
        q_bias_data + k_bias_data + v_bias_data,
    )
    initializers = [
        float_tensor("layernorm_weight", [hidden_size]),
        float_tensor("layernorm_bias", [hidden_size]),
        float_tensor("matmul_after_attn_initializer", [hidden_size, hidden_size]),
        float_tensor("add_after_attn_initializer", [hidden_size]),
    ]

    if fused:
        # Add packed QKV weight tensor as initializer
        initializers.append(qkv_weight)

        # Add Slice indices as initializers
        initializers.extend(
            [
                helper.make_tensor(name="MatMul_0_q_start_index", data_type=TensorProto.INT64, dims=[1], vals=[0]),
                helper.make_tensor(
                    name="MatMul_0_k_start_index", data_type=TensorProto.INT64, dims=[1], vals=[hidden_size]
                ),
                helper.make_tensor(
                    name="MatMul_0_v_start_index", data_type=TensorProto.INT64, dims=[1], vals=[2 * hidden_size]
                ),
                helper.make_tensor(
                    name="MatMul_0_end_of_qkv_index", data_type=TensorProto.INT64, dims=[1], vals=[3 * hidden_size]
                ),
                helper.make_tensor(name="MatMul_0_qkv_last_axis", data_type=TensorProto.INT64, dims=[1], vals=[-1]),
            ]
        )

        # Add packed QKV bias tensor as initializer
        initializers.append(qkv_bias)
    else:
        # Add Q/K/V weight tensors as initializers
        initializers.extend([q_weight, k_weight, v_weight])

        if add_k:
            initializers.extend([q_bias, k_bias, v_bias])
        else:
            initializers.extend([q_bias, v_bias])

        initializers.extend(
            [
                numpy_helper.from_array(np.array(num_heads, dtype="int64"), name="num_heads_int"),
                numpy_helper.from_array(np.array([num_heads], dtype="int64"), name="num_heads"),
                numpy_helper.from_array(np.array([head_size], dtype="int64"), name="head_size"),
                numpy_helper.from_array(np.array([hidden_size], dtype="int64"), name="hidden_size"),
                numpy_helper.from_array(np.array(1 / np.sqrt(head_size), dtype="float32"), name="q_scale"),
                numpy_helper.from_array(np.array(0, dtype="int64"), name="idx_0"),
                numpy_helper.from_array(np.array(1, dtype="int64"), name="idx_1"),
                numpy_helper.from_array(np.array([-1], dtype="int64"), name="neg_one"),
                numpy_helper.from_array(np.array([0], dtype="int64"), name="unsqueeze_axes_input"),
            ]
        )

    # Construct graph
    graph = helper.make_graph(
        nodes, "whisper_decoder_with_past_self_mha_graph", inputs, outputs, initializers, doc_string="whisper"
    )
    opsetid = helper.make_opsetid("ai.onnx", min(onnx.defs.onnx_opset_version(), 16))
    return helper.make_model(graph, opset_imports=(opsetid,))


def create_whisper_decoder_with_past_multihead_cross_attention(
    hidden_size=768, num_heads=12, epsilon=0.000009999999747378752, fused=False
):
    # Get head size and ensure head size is an integer
    assert hidden_size % num_heads == 0
    head_size = hidden_size // num_heads

    # Construct input and output nodes
    inputs = [
        helper.make_tensor_value_info("input_0", TensorProto.FLOAT, ["batch_size", 1500, hidden_size]),
        helper.make_tensor_value_info(
            "past_key_values.0.encoder.key", TensorProto.FLOAT, ["batch_size", num_heads, "past_seq_len", head_size]
        ),
        helper.make_tensor_value_info(
            "past_key_values.0.encoder.value", TensorProto.FLOAT, ["batch_size", num_heads, "past_seq_len", head_size]
        ),
    ]
    outputs = [
        helper.make_tensor_value_info("output_0", TensorProto.FLOAT, ["batch_size", 1500, hidden_size]),
        helper.make_tensor_value_info("output_1", TensorProto.FLOAT, ["batch_size", "sequence_length", hidden_size]),
    ]

    # Create SkipLayerNorm (since there's no Add + LayerNorm variant for this attention subgraph)
    nodes = [
        helper.make_node(
            "SkipLayerNormalization",
            ["input_0", "input_0", "layernorm_weight", "layernorm_bias"],
            ["layernorm_output_to_matmul", "", "", "layernorm_output_to_skiplayernorm"],
            "skiplayernorm",
            domain="com.microsoft",
            epsilon=epsilon,
        )
    ]

    if fused:
        nodes.extend(
            [
                helper.make_node(
                    "MatMul", ["layernorm_output_to_matmul", "q_weight"], ["q_matmul_output"], "q_path_matmul"
                ),
                helper.make_node(
                    "MultiHeadAttention",
                    [
                        "q_matmul_output",
                        "past_key_values.0.encoder.key",
                        "past_key_values.0.encoder.value",
                        "Attention_0_qkv_bias",
                    ],
                    ["attn_output"],
                    "Attention_0",
                    domain="com.microsoft",
                    num_heads=num_heads,
                ),
            ]
        )
    else:
        # Create nodes for Q/K/V paths
        q_nodes = [
            helper.make_node(
                "MatMul", ["layernorm_output_to_matmul", "q_weight"], ["q_matmul_output"], "q_path_matmul"
            ),
            helper.make_node("Add", ["q_bias", "q_matmul_output"], ["q_add_output"], "q_path_add"),
            helper.make_node("Mul", ["q_add_output", "q_scale"], ["q_mul_output"], "q_path_mul"),
            helper.make_node("Reshape", ["q_mul_output", "q_bsnh_reshape"], ["q_4d_bsnh"], "q_reshape_to_4d"),
            helper.make_node("Transpose", ["q_4d_bsnh"], ["q_4d_bnsh"], "q_transpose_to_bnsh", perm=[0, 2, 1, 3]),
            helper.make_node(
                "Reshape",
                ["q_4d_bnsh", "q_attn_heads_output"],
                ["q_output_(num_heads*batch_size,seq_len,head_size)"],
                "q_reshape_to_3d",
            ),
        ]
        k_nodes = [
            helper.make_node(
                "Reshape",
                ["past_key_values.0.encoder.key", "k_attn_heads_output"],
                ["k_output_(num_heads*batch_size,seq_len,head_size)"],
                "k_reshape_to_3d",
            ),
            helper.make_node(
                "Transpose",
                ["k_output_(num_heads*batch_size,seq_len,head_size)"],
                ["k_output_(num_heads*batch_size,head_size,seq_len)"],
                "k_transpose_last_two_dims",
                perm=[0, 2, 1],
            ),
        ]
        v_nodes = [
            helper.make_node(
                "Reshape",
                ["past_key_values.0.encoder.value", "v_attn_heads_output"],
                ["v_output_(num_heads*batch_size,seq_len,head_size)"],
                "v_reshape_to_3d",
            ),
        ]
        nodes.extend(q_nodes)
        nodes.extend(k_nodes)
        nodes.extend(v_nodes)

        # Create nodes used with qkv concats, reshapes, and transposes
        nodes.extend(
            [
                helper.make_node("Shape", ["layernorm_output_to_matmul"], ["shape_output"], "shape"),
                helper.make_node("Gather", ["shape_output", "idx_0"], ["gather_0_output"], "gather_0", axis=0),
                helper.make_node(
                    "Mul", ["gather_0_output", "num_heads_int"], ["mul_attn_heads_output"], "mul_num_heads"
                ),
                helper.make_node(
                    "Unsqueeze",
                    ["mul_attn_heads_output", "unsqueeze_axes_input"],
                    ["unsqueeze_attn_heads_output"],
                    "unsqueeze_num_heads",
                ),
                helper.make_node(
                    "Concat",
                    ["unsqueeze_attn_heads_output", "neg_one", "head_size"],
                    ["q_attn_heads_output"],
                    "q_num_heads",
                    axis=0,
                ),
                helper.make_node(
                    "Concat",
                    ["unsqueeze_attn_heads_output", "neg_one", "head_size"],
                    ["k_attn_heads_output"],
                    "k_num_heads",
                    axis=0,
                ),
                helper.make_node(
                    "Concat",
                    ["unsqueeze_attn_heads_output", "neg_one", "head_size"],
                    ["v_attn_heads_output"],
                    "v_num_heads",
                    axis=0,
                ),
                helper.make_node(
                    "Constant",
                    inputs=[],
                    outputs=["q_bsnh_reshape"],
                    value=numpy_helper.from_array(
                        np.array([0, 0, num_heads, head_size], dtype="int64"), name="const_tensor"
                    ),
                ),
            ]
        )

        # Create nodes used with Q x K' and softmax(Q x K') x V
        nodes.extend(
            [
                helper.make_node("Gather", ["shape_output", "idx_1"], ["gather_1_output"], "gather_1", axis=0),
                helper.make_node(
                    "Unsqueeze", ["gather_0_output", "unsqueeze_axes_input"], ["unsqueeze_0_output"], "unsqueeze_0"
                ),
                helper.make_node(
                    "Unsqueeze", ["gather_1_output", "unsqueeze_axes_input"], ["unsqueeze_1_output"], "unsqueeze_1"
                ),
                helper.make_node(
                    "Concat",
                    ["unsqueeze_0_output", "num_heads", "unsqueeze_1_output", "head_size"],
                    ["bnsh_format"],
                    axis=0,
                ),
                helper.make_node(
                    "Concat",
                    ["unsqueeze_0_output", "unsqueeze_1_output", "hidden_size"],
                    ["bsd_format"],
                    axis=0,
                ),
            ]
        )

        # Create nodes for computing softmax(Q x K') x V
        nodes.extend(
            [
                helper.make_node(
                    "MatMul",
                    [
                        "q_output_(num_heads*batch_size,seq_len,head_size)",
                        "k_output_(num_heads*batch_size,head_size,seq_len)",
                    ],
                    ["qk_output_(num_heads*batch_size,seq_len,seq_len)"],
                    "matmul_qk",
                ),
                helper.make_node(
                    "Softmax",
                    ["qk_output_(num_heads*batch_size,seq_len,seq_len)"],
                    ["softmax_output"],
                    "softmax_qk",
                    axis=2,
                ),
                helper.make_node(
                    "MatMul",
                    ["softmax_output", "v_output_(num_heads*batch_size,seq_len,head_size)"],
                    ["qkv_output_(num_heads*batch_size,seq_len,head_size)"],
                    "matmul_qkv",
                ),
                helper.make_node(
                    "Reshape",
                    ["qkv_output_(num_heads*batch_size,seq_len,head_size)", "bnsh_format"],
                    ["qkv_bnsh"],
                    "reshape_qkv_to_bnsh",
                ),
                helper.make_node("Transpose", ["qkv_bnsh"], ["qkv_bsnh"], "transpose_bnsh_to_bsnh", perm=[0, 2, 1, 3]),
                helper.make_node("Reshape", ["qkv_bsnh", "bsd_format"], ["attn_output"], "qkv_bsd"),
            ]
        )

    # Create final nodes to conclude attention
    nodes.append(
        helper.make_node(
            "MatMul",
            ["attn_output", "matmul_after_attn_initializer"],
            ["matmul_after_attn_output"],
            "matmul_after_attn",
        ),
    )
    if not fused:
        next_sln_inputs = [
            "layernorm_output_to_skiplayernorm",
            "add_after_attn_output",
            "layernorm_weight",
            "layernorm_bias",
        ]
        nodes.extend(
            [
                helper.make_node(
                    "Add",
                    ["add_after_attn_initializer", "matmul_after_attn_output"],
                    ["add_after_attn_output"],
                    "add_after_attn",
                ),
                helper.make_node(
                    "SkipLayerNormalization",
                    next_sln_inputs,
                    ["output_0", "", "", "output_1"],
                    "next_skiplayernorm",
                    domain="com.microsoft",
                    epsilon=epsilon,
                ),
            ]
        )
    else:
        next_sln_inputs = [
            "matmul_after_attn_output",
            "layernorm_output_to_skiplayernorm",
            "layernorm_weight",
            "layernorm_bias",
            "add_after_attn_initializer",
        ]
        nodes.append(
            helper.make_node(
                "SkipLayerNormalization",
                next_sln_inputs,
                ["output_0", "", "", "output_1"],
                "SkipLayerNorm_AddBias_0",
                domain="com.microsoft",
                epsilon=epsilon,
            )
        )

    # Create initializers
    q_weight, q_weight_data = get_tensor_and_weight("q_weight", [hidden_size, hidden_size])
    q_bias, q_bias_data = get_tensor_and_weight("q_bias", [hidden_size])
    k_bias, k_bias_data = get_tensor_and_weight("k_bias", [hidden_size], zeros=True)
    v_bias, v_bias_data = get_tensor_and_weight("v_bias", [hidden_size], zeros=True)
    qkv_bias = helper.make_tensor(
        "Attention_0_qkv_bias", TensorProto.FLOAT, [3 * hidden_size], q_bias_data + k_bias_data + v_bias_data
    )
    initializers = [
        float_tensor("layernorm_weight", [hidden_size]),
        float_tensor("layernorm_bias", [hidden_size]),
        float_tensor("matmul_after_attn_initializer", [hidden_size, hidden_size]),
        float_tensor("add_after_attn_initializer", [hidden_size]),
        q_weight,
    ]

    if fused:
        # Add packed QKV bias tensor as initializer
        initializers.append(qkv_bias)
    else:
        # Add Q bias tensor as initializer
        initializers.append(q_bias)

        initializers.extend(
            [
                numpy_helper.from_array(np.array(num_heads, dtype="int64"), name="num_heads_int"),
                numpy_helper.from_array(np.array([num_heads], dtype="int64"), name="num_heads"),
                numpy_helper.from_array(np.array([head_size], dtype="int64"), name="head_size"),
                numpy_helper.from_array(np.array([hidden_size], dtype="int64"), name="hidden_size"),
                numpy_helper.from_array(np.array(1 / np.sqrt(head_size), dtype="float32"), name="q_scale"),
                numpy_helper.from_array(np.array(0, dtype="int64"), name="idx_0"),
                numpy_helper.from_array(np.array(1, dtype="int64"), name="idx_1"),
                numpy_helper.from_array(np.array([-1], dtype="int64"), name="neg_one"),
                numpy_helper.from_array(np.array([0], dtype="int64"), name="unsqueeze_axes_input"),
            ]
        )

    # Construct graph
    graph = helper.make_graph(
        nodes, "whisper_decoder_with_past_cross_mha_graph", inputs, outputs, initializers, doc_string="whisper"
    )
    opsetid = helper.make_opsetid("ai.onnx", min(onnx.defs.onnx_opset_version(), 16))
    return helper.make_model(graph, opset_imports=(opsetid,))


if __name__ == "__main__":
    np.random.seed(2)
    num_heads = 4
    hidden_size = 64

    model = create_whisper_encoder_attention(num_heads=num_heads, hidden_size=hidden_size)
    onnx.save(model, "whisper_encoder_attention_sln.onnx")

    model = create_whisper_encoder_attention(num_heads=num_heads, hidden_size=hidden_size, fused=True)
    onnx.save(model, "./test_data/models/whisper/encoder_attention_with_sln_fused.onnx")

    model = create_whisper_decoder_attention(num_heads=num_heads, hidden_size=hidden_size)
    onnx.save(model, "whisper_decoder_attention_sln.onnx")

    model = create_whisper_decoder_attention(num_heads=num_heads, hidden_size=hidden_size, fused=True)
    onnx.save(model, "./test_data/models/whisper/decoder_attention_with_sln_fused.onnx")

    model = create_whisper_decoder_multihead_attention(num_heads=num_heads, hidden_size=hidden_size)
    onnx.save(model, "whisper_decoder_mha.onnx")

    model = create_whisper_decoder_multihead_attention(num_heads=num_heads, hidden_size=hidden_size, fused=True)
    onnx.save(model, "./test_data/models/whisper/decoder_mha_fused.onnx")

    model = create_whisper_decoder_with_past_multihead_self_attention(num_heads=num_heads, hidden_size=hidden_size)
    onnx.save(model, "whisper_decoder_with_past_self_mha.onnx")

    model = create_whisper_decoder_with_past_multihead_self_attention(
        num_heads=num_heads, hidden_size=hidden_size, fused=True
    )
    onnx.save(model, "./test_data/models/whisper/decoder_with_past_self_mha_fused.onnx")

    model = create_whisper_decoder_with_past_multihead_cross_attention(num_heads=num_heads, hidden_size=hidden_size)
    onnx.save(model, "whisper_decoder_with_past_cross_mha.onnx")

    model = create_whisper_decoder_with_past_multihead_cross_attention(
        num_heads=num_heads, hidden_size=hidden_size, fused=True
    )
    onnx.save(model, "./test_data/models/whisper/decoder_with_past_cross_mha_fused.onnx")



import argparse
import datetime
import logging
import os
import random
import time
import zipfile

import numpy as np
import pandas as pd
import torch
import wget
from sklearn.model_selection import train_test_split
from torch.utils.data import DataLoader, RandomSampler, SequentialSampler, TensorDataset
from transformers import (
    AdamW,  # noqa: F401
    AutoConfig,
    BertConfig,  # noqa: F401
    BertForSequenceClassification,
    BertTokenizer,
    get_linear_schedule_with_warmup,
)

import onnxruntime
from onnxruntime.training.ortmodule import DebugOptions, ORTModule


def train(model, optimizer, scaler, scheduler, train_dataloader, epoch, device, args):
    # ========================================
    #               Training
    # ========================================
    # This training code is based on the `run_glue.py` script here:
    # https://github.com/huggingface/transformers/blob/5bfcd0485ece086ebcbed2d008813037968a9e58/examples/run_glue.py#L128

    # Perform one full pass over the training set.
    print(f"\n======== Epoch {epoch + 1} / {args.epochs} with batch size {args.batch_size} ========")

    # Measure how long the training epoch takes.
    t0 = time.time()
    start_time = t0

    # Reset the total loss for this epoch.
    total_loss = 0

    # Put the model into training mode. Don't be mislead--the call to
    # `train` just changes the *mode*, it doesn't *perform* the training.
    # `dropout` and `batchnorm` layers behave differently during training
    # vs. test (source: https://stackoverflow.com/questions/51433378/what-does-model-train-do-in-pytorch)
    model.train()

    # Always clear any previously calculated gradients before performing a
    # backward pass. PyTorch doesn't do this automatically because
    # accumulating the gradients is "convenient while training RNNs".
    # (source: https://stackoverflow.com/questions/48001598/why-do-we-need-to-call-zero-grad-in-pytorch)
    optimizer.zero_grad()

    # For each batch of training data...
    for step, batch in enumerate(train_dataloader):
        if step == args.train_steps:
            break

        # Unpack this training batch from our dataloader.
        #
        # As we unpack the batch, we'll also copy each tensor to the GPU using the
        # `to` method.
        #
        # `batch` contains three pytorch tensors:
        #   [0]: input ids
        #   [1]: attention masks
        #   [2]: labels
        b_input_ids = batch[0].to(device)
        b_input_mask = batch[1].to(device)
        b_labels = batch[2].to(device)

        # Perform a forward pass (evaluate the model on this training batch).
        # This will return the loss (rather than the model output) because we have provided the `labels`.
        # The documentation for this `model` function is here:
        #   https://huggingface.co/transformers/v2.2.0/model_doc/bert.html#transformers.BertForSequenceClassification
        with torch.cuda.amp.autocast():
            outputs = model(b_input_ids, attention_mask=b_input_mask, labels=b_labels)

            # The call to `model` always returns a tuple, so we need to pull the
            # loss value out of the tuple.
            loss = outputs[0]

        # Progress update every 40 batches.
        if step % args.log_interval == 0 and step != 0:
            # Calculate elapsed time in minutes.
            curr_time = time.time()
            elapsed_time = curr_time - start_time

            # Report progress.
            print(
                f"Batch {step:4} of {len(train_dataloader):4}. Execution time: {elapsed_time:.4f}. Loss: {loss.item():.4f}"
            )
            start_time = curr_time

        if args.view_graphs:
            import torchviz

            pytorch_backward_graph = torchviz.make_dot(outputs[0], params=dict(list(model.named_parameters())))
            pytorch_backward_graph.view()

        # Accumulate the training loss over all of the batches so that we can
        # calculate the average loss at the end. `loss` is a Tensor containing a
        # single value; the `.item()` function just returns the Python value
        # from the tensor.
        total_loss += loss.item()

        # Perform a backward pass to calculate the gradients.
        scaler.scale(loss).backward()

        # Clip the norm of the gradients to 1.0.
        # This is to help prevent the "exploding gradients" problem.
        torch.nn.utils.clip_grad_norm_(model.parameters(), 1.0)

        if step % args.grad_acc_steps == 0:
            # Update parameters and take a step using the computed gradient.
            # The optimizer dictates the "update rule"--how the parameters are
            # modified based on their gradients, the learning rate, etc.
            scaler.step(optimizer)
            scaler.update()

            # Update the learning rate.
            scheduler.step()
            optimizer.zero_grad()

    # Calculate the average loss over the training data.
    avg_train_loss = total_loss / len(train_dataloader)

    epoch_time = time.time() - t0
    print(f"\n  Average training loss: {avg_train_loss:.2f}")
    print(f"  Training epoch took: {epoch_time:.4f}s")
    return epoch_time


def test(model, validation_dataloader, device, args):
    # ========================================
    #               Validation
    # ========================================
    # After the completion of each training epoch, measure our performance on
    # our validation set.
    print(f"\nRunning Validation with batch size {args.test_batch_size} ...")

    # Put the model in evaluation mode--the dropout layers behave differently
    # during evaluation.
    model.eval()

    t0 = time.time()

    # Tracking variables
    eval_loss, eval_accuracy = 0, 0  # noqa: F841
    nb_eval_steps, nb_eval_examples = 0, 0  # noqa: F841

    # Evaluate data for one epoch
    for batch in validation_dataloader:
        # Add batch to GPU
        batch = tuple(t.to(device) for t in batch)  # noqa: PLW2901

        # Unpack the inputs from our dataloader
        b_input_ids, b_input_mask, b_labels = batch
        # Telling the model not to compute or store gradients, saving memory and
        # speeding up validation
        with torch.no_grad():
            # Forward pass, calculate logit predictions.
            # This will return the logits rather than the loss because we have
            # not provided labels.
            # token_type_ids is the same as the "segment ids", which
            # differentiates sentence 1 and 2 in 2-sentence tasks.
            # The documentation for this `model` function is here:
            # https://huggingface.co/transformers/v2.2.0/model_doc/bert.html#transformers.BertForSequenceClassification

            # TODO: original sample had the last argument equal to None, but b_labels is because model was
            #       exported using 3 inputs for training, so validation must follow.
            #       Another approach would be checkpoint the trained model, re-export the model for validation with the checkpoint
            outputs = model(b_input_ids, attention_mask=b_input_mask, labels=b_labels)

        # Get the "logits" output by the model. The "logits" are the output
        # values prior to applying an activation function like the softmax.
        logits = outputs[1]

        # Move logits and labels to CPU
        logits = logits.detach().cpu().numpy()
        label_ids = b_labels.to("cpu").numpy()

        # Calculate the accuracy for this batch of test sentences.
        tmp_eval_accuracy = flat_accuracy(logits, label_ids)

        # Accumulate the total accuracy.
        eval_accuracy += tmp_eval_accuracy

        # Track the number of batches
        nb_eval_steps += 1

    # Report the final accuracy for this validation run.
    epoch_time = time.time() - t0
    accuracy = eval_accuracy / nb_eval_steps
    print(f"  Accuracy: {accuracy:.2f}")
    print(f"  Validation took: {epoch_time:.4f}s")
    return epoch_time, accuracy


def load_dataset(args):
    # 2. Loading CoLA Dataset

    def _download_dataset(download_dir):
        if not os.path.exists(download_dir):
            # Download the file (if we haven't already)
            print("Downloading dataset...")
            url = "https://nyu-mll.github.io/CoLA/cola_public_1.1.zip"
            wget.download(url, "./cola_public_1.1.zip")
        else:
            print("Reusing cached dataset")

    if not os.path.exists(args.data_dir):
        _download_dataset("./cola_public_1.1.zip")
        # Unzip it
        print("Extracting dataset")
        with zipfile.ZipFile("./cola_public_1.1.zip", "r") as zip_ref:
            zip_ref.extractall("./")
    else:
        print("Reusing extracted dataset")

    # Load the dataset into a pandas dataframe.
    df = pd.read_csv(
        os.path.join(args.data_dir if os.path.exists(args.data_dir) else "cola_public/raw", "in_domain_train.tsv"),
        delimiter="\t",
        header=None,
        names=["sentence_source", "label", "label_notes", "sentence"],
    )

    # Get the lists of sentences and their labels.
    sentences = df.sentence.values
    labels = df.label.values

    # 3. Tokenization & Input Formatting

    # Load the BERT tokenizer.
    tokenizer = BertTokenizer.from_pretrained("bert-base-uncased", do_lower_case=True)

    # Set the max length of encoded sentence.
    # 64 is slightly larger than the maximum training sentence length of 47...
    MAX_LEN = 64  # noqa: N806

    # Tokenize all of the sentences and map the tokens to their word IDs.
    input_ids = []
    for sent in sentences:
        # `encode` will:
        #   (1) Tokenize the sentence.
        #   (2) Prepend the `[CLS]` token to the start.
        #   (3) Append the `[SEP]` token to the end.
        #   (4) Map tokens to their IDs.
        encoded_sent = tokenizer.encode(
            sent,  # Sentence to encode.
            add_special_tokens=True,  # Add '[CLS]' and '[SEP]'
        )

        # Pad our input tokens with value 0.
        if len(encoded_sent) < MAX_LEN:
            encoded_sent.extend([0] * (MAX_LEN - len(encoded_sent)))

        # Truncate to MAX_LEN
        if len(encoded_sent) > MAX_LEN:
            encoded_sent = encoded_sent[:MAX_LEN]

        # Add the encoded sentence to the list.
        input_ids.append(encoded_sent)

    input_ids = np.array(input_ids, dtype=np.longlong)

    # Create attention masks
    attention_masks = []

    # For each sentence...
    for sent in input_ids:
        # Create the attention mask.
        #   - If a token ID is 0, then it's padding, set the mask to 0.
        #   - If a token ID is > 0, then it's a real token, set the mask to 1.
        att_mask = [int(token_id > 0) for token_id in sent]

        # Store the attention mask for this sentence.
        attention_masks.append(att_mask)

    # Use 90% for training and 10% for validation.
    train_inputs, validation_inputs, train_labels, validation_labels = train_test_split(
        input_ids, labels, random_state=2018, test_size=0.1
    )
    # Do the same for the masks.
    train_masks, validation_masks, _, _ = train_test_split(attention_masks, labels, random_state=2018, test_size=0.1)

    # Convert all inputs and labels into torch tensors, the required datatype
    # for our model.
    train_inputs = torch.tensor(train_inputs)
    validation_inputs = torch.tensor(validation_inputs)

    train_labels = torch.tensor(train_labels)
    validation_labels = torch.tensor(validation_labels)

    train_masks = torch.tensor(train_masks)
    validation_masks = torch.tensor(validation_masks)

    # Create the DataLoader for our training set.
    train_data = TensorDataset(train_inputs, train_masks, train_labels)
    train_sampler = RandomSampler(train_data)
    train_dataloader = DataLoader(train_data, sampler=train_sampler, batch_size=args.batch_size)

    # Create the DataLoader for our validation set.
    validation_data = TensorDataset(validation_inputs, validation_masks, validation_labels)
    validation_sampler = SequentialSampler(validation_data)
    validation_dataloader = DataLoader(validation_data, sampler=validation_sampler, batch_size=args.test_batch_size)

    return train_dataloader, validation_dataloader


# Function to calculate the accuracy of our predictions vs labels
def flat_accuracy(preds, labels):
    pred_flat = np.argmax(preds, axis=1).flatten()
    labels_flat = labels.flatten()
    return np.sum(pred_flat == labels_flat) / len(labels_flat)


def format_time(elapsed):
    """Takes a time in seconds and returns a string hh:mm:ss"""
    # Round to the nearest second.
    elapsed_rounded = int(round(elapsed))

    # Format as hh:mm:ss
    return str(datetime.timedelta(seconds=elapsed_rounded))


def main():
    # 1. Basic setup
    parser = argparse.ArgumentParser(description="PyTorch MNIST Example")
    parser.add_argument("--pytorch-only", action="store_true", default=False, help="disables ONNX Runtime training")
    parser.add_argument(
        "--batch-size", type=int, default=32, metavar="N", help="input batch size for training (default: 32)"
    )
    parser.add_argument("--do-val", action="store_true", default=False, help="disables validation")
    parser.add_argument(
        "--test-batch-size", type=int, default=64, metavar="N", help="input batch size for testing (default: 64)"
    )
    parser.add_argument("--view-graphs", action="store_true", default=False, help="views forward and backward graphs")
    parser.add_argument(
        "--export-onnx-graphs", action="store_true", default=False, help="export ONNX graphs to current directory"
    )
    parser.add_argument("--no-cuda", action="store_true", default=False, help="disables CUDA training")
    parser.add_argument("--epochs", type=int, default=4, metavar="N", help="number of epochs to train (default: 4)")
    parser.add_argument("--seed", type=int, default=42, metavar="S", help="random seed (default: 42)")
    parser.add_argument(
        "--log-interval",
        type=int,
        default=40,
        metavar="N",
        help="how many batches to wait before logging training status (default: 40)",
    )
    parser.add_argument(
        "--train-steps",
        type=int,
        default=-1,
        metavar="N",
        help="number of steps to train. Set -1 to run through whole dataset (default: -1)",
    )
    parser.add_argument(
        "--log-level",
        choices=["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
        default="WARNING",
        help="Log level (default: WARNING)",
    )
    parser.add_argument(
        "--num-hidden-layers",
        type=int,
        default=1,
        metavar="H",
        help="Number of hidden layers for the BERT model. A vanila BERT has 12 hidden layers (default: 1)",
    )
    parser.add_argument("--data-dir", type=str, default="./cola_public/raw", help="Path to the bert data directory")
    parser.add_argument("--grad-acc-steps", type=int, default=2, help="Number of steps for accumulating gradients")

    args = parser.parse_args()

    # Device (CPU vs CUDA)
    if torch.cuda.is_available() and not args.no_cuda:
        device = torch.device("cuda")
        print(f"There are {torch.cuda.device_count()} GPU(s) available.")
        print("We will use the GPU:", torch.cuda.get_device_name(0))
    else:
        print("No GPU available, using the CPU instead.")
        device = torch.device("cpu")

    # Set log level
    numeric_level = getattr(logging, args.log_level.upper(), None)
    if not isinstance(numeric_level, int):
        raise ValueError(f"Invalid log level: {args.log_level}")
    logging.basicConfig(level=numeric_level)

    # 2. Dataloader
    train_dataloader, validation_dataloader = load_dataset(args)

    # 3. Modeling
    # Load BertForSequenceClassification, the pretrained BERT model with a single
    # linear classification layer on top.
    config = AutoConfig.from_pretrained(
        "bert-base-uncased",
        num_labels=2,
        num_hidden_layers=args.num_hidden_layers,
        output_attentions=False,  # Whether the model returns attentions weights.
        output_hidden_states=False,  # Whether the model returns all hidden-states.
    )
    model = BertForSequenceClassification.from_pretrained(
        "bert-base-uncased",  # Use the 12-layer BERT model, with an uncased vocab.
        config=config,
    )

    # Note: AdamW is a class from the huggingface library (as opposed to pytorch)
    optimizer = torch.optim.AdamW(
        model.parameters(),
        lr=2e-2,  # args.learning_rate - default is 5e-5, our notebook had 2e-5
        eps=1e-8,  # args.adam_epsilon  - default is 1e-8.
    )

    if not args.pytorch_only:
        # Just for future debugging
        debug_options = DebugOptions(
            save_onnx=args.export_onnx_graphs, onnx_prefix="BertForSequenceClassificationAutoCast"
        )

        model = ORTModule(model, debug_options)
        model._torch_module._execution_manager(is_training=True)._runtime_options.enable_grad_acc_optimization = True

    # Tell pytorch to run this model on the GPU.
    if torch.cuda.is_available() and not args.no_cuda:
        model.cuda()

    # Authors recommend between 2 and 4 epochs
    # Total number of training steps is number of batches * number of epochs.
    total_steps = len(train_dataloader) * args.epochs

    # Create the learning rate scheduler.
    scheduler = get_linear_schedule_with_warmup(
        optimizer,
        num_warmup_steps=0,
        num_training_steps=total_steps,  # Default value in run_glue.py
    )
    scaler = torch.cuda.amp.GradScaler()

    # Seed
    random.seed(args.seed)
    np.random.seed(args.seed)
    torch.manual_seed(args.seed)
    onnxruntime.set_seed(args.seed)
    if torch.cuda.is_available() and not args.no_cuda:
        torch.cuda.manual_seed_all(args.seed)

    # 4. Train loop (fine-tune)
    total_training_time, total_test_time, epoch_0_training, validation_accuracy = 0, 0, 0, 0
    for epoch_i in range(args.epochs):
        total_training_time += train(model, optimizer, scaler, scheduler, train_dataloader, epoch_i, device, args)
        if not args.pytorch_only and epoch_i == 0:
            epoch_0_training = total_training_time
        if args.do_val:
            test_time, validation_accuracy = test(model, validation_dataloader, device, args)
            total_test_time += test_time

    if args.do_val:
        assert validation_accuracy > 0.5

    print("\n======== Global stats ========")
    if not args.pytorch_only:
        estimated_export = 0
        if args.epochs > 1:
            estimated_export = epoch_0_training - (total_training_time - epoch_0_training) / (args.epochs - 1)
            print(f"  Estimated ONNX export took:               {estimated_export:.4f}s")
        else:
            print("  Estimated ONNX export took:               Estimate available when epochs > 1 only")
        print(f"  Accumulated training without export took: {total_training_time - estimated_export:.4f}s")
    print(f"  Accumulated training took:                {total_training_time:.4f}s")
    print(f"  Accumulated validation took:              {total_test_time:.4f}s")


if __name__ == "__main__":
    main()



import onnx
from onnx import TensorProto, helper
from packaging import version

if version.parse(onnx.__version__) == version.parse("1.8.0"):
    opset_version = 13
elif version.parse(onnx.__version__) == version.parse("1.6.0"):
    opset_version = 11
else:
    raise RuntimeError("Please pip install onnx==1.8.0 or 1.6.0 before running this script")


def GenerateNodes(model_name, has_cast, suffix=""):  # noqa: N802
    nodes = [  # LayerNorm subgraph
        helper.make_node("Shape", ["input_ids" + suffix], ["shape1_out" + suffix], "shape1" + suffix),
        helper.make_node(
            "Gather",
            ["shape1_out" + suffix, "indices_0"],
            ["gather0_out" + suffix],
            "gather0" + suffix,
        ),
        (
            helper.make_node(
                "Unsqueeze",
                ["gather0_out" + suffix, "axes_0"],
                ["unsqueeze0_out" + suffix],
                "unsqueeze0" + suffix,
            )
            if opset_version == 13
            else helper.make_node(
                "Unsqueeze",
                ["gather0_out" + suffix],
                ["unsqueeze0_out" + suffix],
                "unsqueeze0" + suffix,
                axes=[0],
            )
        ),
        helper.make_node("Shape", ["input_ids" + suffix], ["shape2_out" + suffix], "shape2" + suffix),
        helper.make_node(
            "Gather",
            ["shape2_out" + suffix, "indices_1"],
            ["gather1_out" + suffix],
            "gather1" + suffix,
        ),
        (
            helper.make_node(
                "Unsqueeze",
                ["gather1_out" + suffix, "axes_0"],
                ["unsqueeze1_out" + suffix],
                "unsqueeze1" + suffix,
            )
            if opset_version == 13
            else helper.make_node(
                "Unsqueeze",
                ["gather1_out" + suffix],
                ["unsqueeze1_out" + suffix],
                "unsqueeze1" + suffix,
                axes=[0],
            )
        ),
        helper.make_node(
            "Concat",
            ["unsqueeze0_out" + suffix, "unsqueeze1_out" + suffix],
            ["concat_out" + suffix],
            "concat" + suffix,
            axis=0,
        ),
        helper.make_node(
            "Cast",
            ["gather1_out" + suffix],
            ["cast_out" + suffix],
            "cast" + suffix,
            to=7,
        ),
        helper.make_node(
            "Range",
            [
                "start_0",
                "cast_out" + suffix if has_cast else "gather1_out" + suffix,
                "delta_1",
            ],
            ["range_out" + suffix],
            "range" + suffix,
        ),
        (
            helper.make_node(
                "Unsqueeze",
                ["range_out" + suffix, "axes_0"],
                ["unsqueeze2_out" + suffix],
                "unsqueeze2" + suffix,
            )
            if opset_version == 13
            else helper.make_node(
                "Unsqueeze",
                ["range_out" + suffix],
                ["unsqueeze2_out" + suffix],
                "unsqueeze2" + suffix,
                axes=[0],
            )
        ),
        helper.make_node(
            "Expand",
            ["unsqueeze2_out" + suffix, "concat_out" + suffix],
            ["expand_out" + suffix],
            "expand" + suffix,
        ),
        helper.make_node(
            "Gather",
            ["pos_embed", "expand_out" + suffix],
            ["pos_gather_out" + suffix],
            "pos_gather" + suffix,
        ),
        helper.make_node(
            "Gather",
            ["word_embed", "input_ids" + suffix],
            ["word_gather_out" + suffix],
            "word_gather" + suffix,
        ),
        helper.make_node(
            "Add",
            ["word_gather_out" + suffix, "pos_gather_out" + suffix],
            ["word_add_pos_out" + suffix],
            "word_add_pos" + suffix,
        ),
        helper.make_node(
            "Gather",
            ["seg_embed", "segment_ids" + suffix],
            ["seg_gather_out" + suffix],
            "seg_gather" + suffix,
        ),
        helper.make_node(
            "Add",
            ["word_add_pos_out" + suffix, "seg_gather_out" + suffix],
            ["add3_out" + suffix],
            "add3" + suffix,
        ),
        helper.make_node(
            "LayerNormalization",
            ["add3_out" + suffix, "layer_norm_weight", "layer_norm_bias"],
            ["layernorm_out" + suffix],
            "layernorm" + suffix,
            axis=-1,
            epsion=0.000009999999747378752,
        ),
        helper.make_node(
            "Cast",
            ["input_mask" + suffix],
            ["mask_cast_out" + suffix],
            "mask_cast" + suffix,
            to=6,
        ),
        (
            helper.make_node(
                "ReduceSum",
                ["mask_cast_out" + suffix, "axes_1"],
                ["mask_index_out" + suffix],
                "mask_index" + suffix,
                keepdims=0,
            )
            if opset_version == 13
            else helper.make_node(
                "ReduceSum",
                ["mask_cast_out" + suffix],
                ["mask_index_out" + suffix],
                "mask_index" + suffix,
                axes=[1],
                keepdims=0,
            )
        ),
        helper.make_node(
            "Attention",
            [
                "layernorm_out" + suffix,
                "qkv_weights",
                "qkv_bias",
                "mask_index_out" + suffix,
            ],
            ["att_out" + suffix],
            "att" + suffix,
            domain="com.microsoft",
            num_heads=2,
        ),
        helper.make_node(
            "MatMul",
            ["att_out" + suffix, "matmul_weight"],
            ["matmul_out" + suffix],
            "matmul" + suffix,
        ),
        helper.make_node(
            "Add",
            ["matmul_out" + suffix, "add_bias"],
            ["add_out" + suffix],
            "add" + suffix,
        ),
        helper.make_node(
            "Add",
            ["add_out" + suffix, "layernorm_out" + suffix],
            ["add2_out" + suffix],
            "add2" + suffix,
        ),
    ]

    if not has_cast:
        del nodes[7:8]
    return nodes


def GenerateInitializers():  # noqa: N802
    # hidden_size=4, num_heads=2
    initializers = [  # initializers
        helper.make_tensor("indices_0", TensorProto.INT64, [], [0]),
        helper.make_tensor("indices_1", TensorProto.INT64, [], [1]),
        helper.make_tensor("start_0", TensorProto.INT64, [], [0]),
        helper.make_tensor("delta_1", TensorProto.INT64, [], [1]),
        helper.make_tensor(
            "word_embed",
            TensorProto.FLOAT,
            [2, 4],
            [1.0, 2.0, 3.0, 4.0, 1.0, 2.0, 3.0, 4.0],
        ),
        helper.make_tensor(
            "pos_embed",
            TensorProto.FLOAT,
            [4, 4],
            [
                1.0,
                2.0,
                3.0,
                4.0,
                1.0,
                2.0,
                3.0,
                4.0,
                1.0,
                2.0,
                3.0,
                4.0,
                1.0,
                2.0,
                3.0,
                4.0,
            ],
        ),
        helper.make_tensor(
            "seg_embed",
            TensorProto.FLOAT,
            [2, 4],
            [1.0, 2.0, 3.0, 4.0, 1.0, 2.0, 3.0, 4.0],
        ),
        helper.make_tensor("layer_norm_weight", TensorProto.FLOAT, [4], [1.0, 2.0, 3.0, 4.0]),
        helper.make_tensor("layer_norm_bias", TensorProto.FLOAT, [4], [0.1, 0.2, 0.3, 0.4]),
        helper.make_tensor("qkv_weights", TensorProto.FLOAT, [4, 12], [0.1] * 4 * 12),
        helper.make_tensor(
            "qkv_bias",
            TensorProto.FLOAT,
            [12],
            [0.1, 0.2, 0.3, 0.4, 0.1, 0.2, 0.3, 0.4, 0.1, 0.2, 0.3, 0.4],
        ),
        helper.make_tensor(
            "matmul_weight",
            TensorProto.FLOAT,
            [4, 4],
            [
                1.0,
                2.0,
                3.0,
                4.0,
                1.0,
                2.0,
                3.0,
                4.0,
                1.0,
                2.0,
                3.0,
                4.0,
                1.0,
                2.0,
                3.0,
                4.0,
            ],
        ),
        helper.make_tensor("add_bias", TensorProto.FLOAT, [4], [0.1, 0.2, 0.3, 0.4]),
        helper.make_tensor("axes_0", TensorProto.INT64, [1], [0]),
        helper.make_tensor("axes_1", TensorProto.INT64, [1], [1]),
    ]

    return initializers


def GenerateMultipleEmbedModel(model_name):  # noqa: N802
    nodes_1 = GenerateNodes(model_name, False, "_1")
    nodes_2 = GenerateNodes(model_name, False, "_2")
    nodes = nodes_1 + nodes_2
    nodes.append(helper.make_node("Add", ["add2_out_1", "add2_out_2"], ["add3_out"], "add3"))

    # hidden_size=4, num_heads=2, max_seq_length=3
    initializers = GenerateInitializers()

    graph = helper.make_graph(
        nodes,
        "EmbedLayerNorm_format3",  # name
        [  # inputs
            helper.make_tensor_value_info("input_ids_1", TensorProto.INT64, ["batch", 3]),
            helper.make_tensor_value_info("segment_ids_1", TensorProto.INT64, ["batch", 3]),
            helper.make_tensor_value_info("input_mask_1", TensorProto.INT64, ["batch", 3]),
            helper.make_tensor_value_info("input_ids_2", TensorProto.INT64, ["batch", 3]),
            helper.make_tensor_value_info("segment_ids_2", TensorProto.INT64, ["batch", 3]),
            helper.make_tensor_value_info("input_mask_2", TensorProto.INT64, ["batch", 3]),
        ],
        [  # outputs
            helper.make_tensor_value_info("add3_out", TensorProto.FLOAT, ["batch", 3, 4]),
        ],
        initializers,
    )

    model = helper.make_model(graph)
    onnx.save(model, model_name)


def GenerateModel3(model_name, has_cast):  # noqa: N802
    nodes = GenerateNodes(model_name, has_cast)

    # hidden_size=4, num_heads=2, max_seq_length=3
    initializers = GenerateInitializers()

    graph = helper.make_graph(
        nodes,
        "EmbedLayerNorm_format3",  # name
        [  # inputs
            helper.make_tensor_value_info("input_ids", TensorProto.INT64, ["batch", 3]),
            helper.make_tensor_value_info("segment_ids", TensorProto.INT64, ["batch", 3]),
            helper.make_tensor_value_info("input_mask", TensorProto.INT64, ["batch", 3]),
        ],
        [  # outputs
            helper.make_tensor_value_info("add2_out", TensorProto.FLOAT, ["batch", 3, 4]),
        ],
        initializers,
    )

    model = helper.make_model(graph)
    onnx.save(model, model_name)


def GenerateModel5(model_name):  # noqa: N802
    batch_size = 2
    hidden_size = 4
    attention_heads = 2
    sequence_length = 3

    nodes = [
        helper.make_node(
            "Gather",
            ["word_embed", "input_ids"],
            ["word_gather_out"],
            "word_gather",
            axis=0,
        ),
        helper.make_node(
            "Add",
            ["word_gather_out", "pos_gather_out"],
            ["word_add_pos_out"],
            "word_add_pos",
        ),
        helper.make_node(
            "Gather",
            ["seg_embed", "segment_ids"],
            ["seg_gather_out"],
            "seg_gather",
            axis=0,
        ),
        helper.make_node("Add", ["word_add_pos_out", "seg_gather_out"], ["add3_out"], "add3"),
        helper.make_node(
            "LayerNormalization",
            ["add3_out", "layer_norm_weight", "layer_norm_bias"],
            ["layernorm_out"],
            "layernorm",
            axis=-1,
            epsion=0.000009999999747378752,
        ),
        helper.make_node("Cast", ["input_mask"], ["mask_cast_out"], "mask_cast", to=6),
        (
            helper.make_node(
                "ReduceSum",
                ["mask_cast_out", "axes_1"],
                ["mask_index_out"],
                "mask_index",
                keepdims=0,
            )
            if opset_version == 13
            else helper.make_node(
                "ReduceSum",
                ["mask_cast_out"],
                ["mask_index_out"],
                "mask_index",
                axes=[1],
                keepdims=0,
            )
        ),
        helper.make_node(
            "Attention",
            ["layernorm_out", "qkv_weights", "qkv_bias", "mask_index_out"],
            ["att_out"],
            "att",
            domain="com.microsoft",
            num_heads=attention_heads,
        ),
        helper.make_node("MatMul", ["att_out", "matmul_weight"], ["matmul_out"], "matmul"),
        helper.make_node("Add", ["matmul_out", "add_bias"], ["add_out"], "add"),
        helper.make_node("Add", ["add_out", "layernorm_out"], ["add2_out"], "add2"),
    ]

    qkv_weights = [1.0] * hidden_size * (3 * hidden_size)

    initializers = [  # initializers
        helper.make_tensor(
            "word_embed",
            TensorProto.FLOAT,
            [2, hidden_size],
            [1.0, 2.0, 3.0, 4.0, 1.0, 2.0, 3.0, 4.0],
        ),
        helper.make_tensor(
            "pos_gather_out",
            TensorProto.FLOAT,
            [batch_size, sequence_length, hidden_size],
            [
                1.0,
                2.0,
                3.0,
                4.0,
                5.0,
                6.0,
                7.0,
                8.0,
                9.0,
                8.0,
                7.0,
                6.0,
                1.0,
                2.0,
                3.0,
                4.0,
                5.0,
                6.0,
                7.0,
                8.0,
                9.0,
                8.0,
                7.0,
                6.0,
            ],
        ),
        helper.make_tensor(
            "seg_embed",
            TensorProto.FLOAT,
            [2, hidden_size],
            [1.0, 2.0, 3.0, 4.0, 1.0, 2.0, 3.0, 4.0],
        ),
        helper.make_tensor("layer_norm_weight", TensorProto.FLOAT, [hidden_size], [1.0, 2.0, 3.0, 4.0]),
        helper.make_tensor("layer_norm_bias", TensorProto.FLOAT, [hidden_size], [0.1, 0.2, 0.3, 0.4]),
        helper.make_tensor(
            "qkv_weights",
            TensorProto.FLOAT,
            [hidden_size, 3 * hidden_size],
            qkv_weights,
        ),
        helper.make_tensor(
            "qkv_bias",
            TensorProto.FLOAT,
            [3 * hidden_size],
            [0.1, 0.2, 0.3, 0.4, 0.1, 0.2, 0.3, 0.4, 0.1, 0.2, 0.3, 0.4],
        ),
        helper.make_tensor(
            "matmul_weight",
            TensorProto.FLOAT,
            [hidden_size, hidden_size],
            [
                1.0,
                2.0,
                3.0,
                4.0,
                1.0,
                2.0,
                3.0,
                4.0,
                1.0,
                2.0,
                3.0,
                4.0,
                1.0,
                2.0,
                3.0,
                4.0,
            ],
        ),
        helper.make_tensor("add_bias", TensorProto.FLOAT, [hidden_size], [0.1, 0.2, 0.3, 0.4]),
        helper.make_tensor("axes_1", TensorProto.INT64, [1], [1]),
    ]

    graph = helper.make_graph(
        nodes,
        "EmbedLayerNorm_format5",  # name
        [  # inputs
            helper.make_tensor_value_info("input_ids", TensorProto.INT64, [batch_size, sequence_length]),
            helper.make_tensor_value_info("segment_ids", TensorProto.INT64, [batch_size, sequence_length]),
            helper.make_tensor_value_info("input_mask", TensorProto.INT64, [batch_size, sequence_length]),
        ],
        [  # outputs
            helper.make_tensor_value_info(
                "add2_out",
                TensorProto.FLOAT,
                [batch_size, sequence_length, hidden_size],
            ),
        ],
        initializers,
    )

    model = helper.make_model(graph)
    onnx.save(model, model_name)


def GenerateModel6(model_name):  # noqa: N802
    nodes = [  # LayerNorm subgraph
        helper.make_node("Shape", ["input_ids"], ["shape1_out"], "shape1"),
        helper.make_node("Gather", ["shape1_out", "indices_0"], ["gather0_out"], "gather0"),
        (
            helper.make_node("Unsqueeze", ["gather0_out", "axes_0"], ["unsqueeze0_out"], "unsqueeze0")
            if opset_version == 13
            else helper.make_node("Unsqueeze", ["gather0_out"], ["unsqueeze0_out"], "unsqueeze0", axes=[0])
        ),
        helper.make_node("Shape", ["input_ids"], ["shape2_out"], "shape2"),
        helper.make_node("Gather", ["shape2_out", "indices_1"], ["gather1_out"], "gather1"),
        (
            helper.make_node("Unsqueeze", ["gather1_out", "axes_0"], ["unsqueeze1_out"], "unsqueeze1")
            if opset_version == 13
            else helper.make_node("Unsqueeze", ["gather1_out"], ["unsqueeze1_out"], "unsqueeze1", axes=[0])
        ),
        helper.make_node(
            "Concat",
            ["unsqueeze0_out", "unsqueeze1_out"],
            ["concat_out"],
            "concat",
            axis=0,
        ),
        helper.make_node("Reshape", ["concat_out", "reshape_init"], ["reshape_out"], "reshape"),
        helper.make_node("Equal", ["reshape_out", "equal_init"], ["equal_out"], "equal"),
        helper.make_node("Where", ["equal_out", "where_init", "reshape_out"], ["where_out"], "where"),
        helper.make_node("Range", ["start_0", "gather1_out", "delta_1"], ["range_out"], "range"),
        (
            helper.make_node("Unsqueeze", ["range_out", "axes_0"], ["unsqueeze2_out"], "unsqueeze2")
            if opset_version == 13
            else helper.make_node("Unsqueeze", ["range_out"], ["unsqueeze2_out"], "unsqueeze2", axes=[0])
        ),
        helper.make_node("Expand", ["unsqueeze2_out", "where_out"], ["expand_out"], "expand"),
        helper.make_node("Gather", ["pos_embed", "expand_out"], ["pos_gather_out"], "pos_gather"),
        helper.make_node("Gather", ["word_embed", "input_ids"], ["word_gather_out"], "word_gather"),
        helper.make_node(
            "Add",
            ["word_gather_out", "pos_gather_out"],
            ["word_add_pos_out"],
            "word_add_pos",
        ),
        helper.make_node("Gather", ["seg_embed", "segment_ids"], ["seg_gather_out"], "seg_gather"),
        helper.make_node("Add", ["word_add_pos_out", "seg_gather_out"], ["add3_out"], "add3"),
        helper.make_node(
            "LayerNormalization",
            ["add3_out", "layer_norm_weight", "layer_norm_bias"],
            ["layernorm_out"],
            "layernorm",
            axis=-1,
            epsion=0.000009999999747378752,
        ),
        helper.make_node("Cast", ["input_mask"], ["mask_cast_out"], "mask_cast", to=6),
        (
            helper.make_node(
                "ReduceSum",
                ["mask_cast_out", "axes_1"],
                ["mask_index_out"],
                "mask_index",
                keepdims=0,
            )
            if opset_version == 13
            else helper.make_node(
                "ReduceSum",
                ["mask_cast_out"],
                ["mask_index_out"],
                "mask_index",
                axes=[1],
                keepdims=0,
            )
        ),
        helper.make_node(
            "Attention",
            ["layernorm_out", "qkv_weights", "qkv_bias", "mask_index_out"],
            ["att_out"],
            "att",
            domain="com.microsoft",
            num_heads=2,
        ),
        helper.make_node("MatMul", ["att_out", "matmul_weight"], ["matmul_out"], "matmul"),
        helper.make_node("Add", ["matmul_out", "add_bias"], ["add_out"], "add"),
        helper.make_node("Add", ["add_out", "layernorm_out"], ["add2_out"], "add2"),
    ]

    # hidden_size=4, num_heads=2, max_seq_length=3
    initializers = [  # initializers
        helper.make_tensor("indices_0", TensorProto.INT64, [], [0]),
        helper.make_tensor("indices_1", TensorProto.INT64, [], [1]),
        helper.make_tensor("start_0", TensorProto.INT64, [], [0]),
        helper.make_tensor("delta_1", TensorProto.INT64, [], [1]),
        helper.make_tensor(
            "word_embed",
            TensorProto.FLOAT,
            [2, 4],
            [1.0, 2.0, 3.0, 4.0, 1.0, 2.0, 3.0, 4.0],
        ),
        helper.make_tensor(
            "pos_embed",
            TensorProto.FLOAT,
            [4, 4],
            [
                1.0,
                2.0,
                3.0,
                4.0,
                1.0,
                2.0,
                3.0,
                4.0,
                1.0,
                2.0,
                3.0,
                4.0,
                1.0,
                2.0,
                3.0,
                4.0,
            ],
        ),
        helper.make_tensor(
            "seg_embed",
            TensorProto.FLOAT,
            [2, 4],
            [1.0, 2.0, 3.0, 4.0, 1.0, 2.0, 3.0, 4.0],
        ),
        helper.make_tensor("layer_norm_weight", TensorProto.FLOAT, [4], [1.0, 2.0, 3.0, 4.0]),
        helper.make_tensor("layer_norm_bias", TensorProto.FLOAT, [4], [0.1, 0.2, 0.3, 0.4]),
        helper.make_tensor("qkv_weights", TensorProto.FLOAT, [4, 12], [0.1] * 4 * 12),
        helper.make_tensor("qkv_bias", TensorProto.FLOAT, [12], [0.1] * 12),
        helper.make_tensor(
            "matmul_weight",
            TensorProto.FLOAT,
            [4, 4],
            [
                1.0,
                2.0,
                3.0,
                4.0,
                1.0,
                2.0,
                3.0,
                4.0,
                1.0,
                2.0,
                3.0,
                4.0,
                1.0,
                2.0,
                3.0,
                4.0,
            ],
        ),
        helper.make_tensor("add_bias", TensorProto.FLOAT, [4], [0.1, 0.2, 0.3, 0.4]),
        helper.make_tensor("reshape_init", TensorProto.INT64, [1], [-1]),
        helper.make_tensor("equal_init", TensorProto.INT64, [2], [-1, -1]),
        helper.make_tensor("where_init", TensorProto.INT64, [2], [1, 1]),
        helper.make_tensor("axes_0", TensorProto.INT64, [1], [0]),
        helper.make_tensor("axes_1", TensorProto.INT64, [1], [1]),
    ]

    graph = helper.make_graph(
        nodes,
        "EmbedLayerNorm_format6",  # name
        [  # inputs
            helper.make_tensor_value_info("input_ids", TensorProto.INT64, ["batch", 3]),
            helper.make_tensor_value_info("segment_ids", TensorProto.INT64, ["batch", 3]),
            helper.make_tensor_value_info("input_mask", TensorProto.INT64, ["batch", 3]),
        ],
        [  # outputs
            helper.make_tensor_value_info("add2_out", TensorProto.FLOAT, ["batch", 3, 4]),
        ],
        initializers,
    )

    model = helper.make_model(graph)
    onnx.save(model, model_name)


def GenerateInitializers2(hidden_size):  # noqa: N802
    qkv_weights = [1.0] * hidden_size * (3 * hidden_size)

    initializers = [  # initializers
        helper.make_tensor(
            "word_embed",
            TensorProto.FLOAT,
            [2, hidden_size],
            [1.0, 2.0, 3.0, 4.0, 1.0, 2.0, 3.0, 4.0],
        ),
        helper.make_tensor(
            "pos_embed",
            TensorProto.FLOAT,
            [2, hidden_size],
            [1.0, 2.0, 3.0, 4.0, 1.0, 2.0, 3.0, 4.0],
        ),
        helper.make_tensor("indices_0", TensorProto.INT64, [], [0]),
        helper.make_tensor("indices_1", TensorProto.INT64, [], [1]),
        helper.make_tensor("start", TensorProto.INT64, [], [0]),
        helper.make_tensor("delta", TensorProto.INT64, [], [1]),
        helper.make_tensor("layer_norm_weight", TensorProto.FLOAT, [hidden_size], [1.0, 2.0, 3.0, 4.0]),
        helper.make_tensor("layer_norm_bias", TensorProto.FLOAT, [hidden_size], [0.1, 0.2, 0.3, 0.4]),
        helper.make_tensor(
            "qkv_weights",
            TensorProto.FLOAT,
            [hidden_size, 3 * hidden_size],
            qkv_weights,
        ),
        helper.make_tensor(
            "qkv_bias",
            TensorProto.FLOAT,
            [3 * hidden_size],
            [0.1, 0.2, 0.3, 0.4, 0.1, 0.2, 0.3, 0.4, 0.1, 0.2, 0.3, 0.4],
        ),
        helper.make_tensor(
            "matmul_weight",
            TensorProto.FLOAT,
            [hidden_size, hidden_size],
            [
                1.0,
                2.0,
                3.0,
                4.0,
                1.0,
                2.0,
                3.0,
                4.0,
                1.0,
                2.0,
                3.0,
                4.0,
                1.0,
                2.0,
                3.0,
                4.0,
            ],
        ),
        helper.make_tensor("add_bias", TensorProto.FLOAT, [hidden_size], [0.1, 0.2, 0.3, 0.4]),
        helper.make_tensor("axes_0", TensorProto.INT64, [1], [0]),
        helper.make_tensor("axes_1", TensorProto.INT64, [1], [1]),
    ]

    return initializers


def GenerateNodes2(attention_heads):  # noqa: N802
    nodes = [
        helper.make_node(
            "Gather",
            ["word_embed", "input_ids"],
            ["word_gather_out"],
            "word_gather",
            axis=0,
        ),
        helper.make_node("Shape", ["input_ids"], ["shape0_out"], "shape0"),
        helper.make_node("Gather", ["shape0_out", "indices_1"], ["gather0_out"], "gather0"),
        helper.make_node("Range", ["start", "gather0_out", "delta"], ["range0_out"], "range0"),
        (
            helper.make_node("Unsqueeze", ["range0_out", "axes_0"], ["unsqueeze0_out"], "unsqueeze0")
            if opset_version == 13
            else helper.make_node("Unsqueeze", ["range0_out"], ["unsqueeze0_out"], "unsqueeze0", axes=[0])
        ),
        helper.make_node("Shape", ["input_ids"], ["shape1_out"], "shape1"),
        helper.make_node("Expand", ["unsqueeze0_out", "shape1_out"], ["expand_out"], "expand"),
        helper.make_node(
            "Gather",
            ["pos_embed", "expand_out"],
            ["pos_gather_out"],
            "pos_gather",
            axis=0,
        ),
        helper.make_node("Add", ["word_gather_out", "pos_gather_out"], ["add1_out"], "add1"),
        helper.make_node(
            "LayerNormalization",
            ["add1_out", "layer_norm_weight", "layer_norm_bias"],
            ["layernorm_out"],
            "layernorm",
            axis=-1,
            epsion=0.000009999999747378752,
        ),
        helper.make_node("Cast", ["input_mask"], ["mask_cast_out"], "mask_cast", to=6),
        (
            helper.make_node(
                "ReduceSum",
                ["mask_cast_out", "axes_1"],
                ["mask_index_out"],
                "mask_index",
                keepdims=0,
            )
            if opset_version == 13
            else helper.make_node(
                "ReduceSum",
                ["mask_cast_out"],
                ["mask_index_out"],
                "mask_index",
                axes=[1],
                keepdims=0,
            )
        ),
        helper.make_node(
            "Attention",
            ["layernorm_out", "qkv_weights", "qkv_bias", "mask_index_out"],
            ["att_out"],
            "att",
            domain="com.microsoft",
            num_heads=attention_heads,
        ),
        helper.make_node("MatMul", ["att_out", "matmul_weight"], ["matmul_out"], "matmul"),
        helper.make_node("Add", ["matmul_out", "add_bias"], ["add2_out"], "add2"),
        helper.make_node("Add", ["add2_out", "layernorm_out"], ["add3_out"], "add3"),
    ]

    return nodes


def GenerateModel7(model_name):  # noqa: N802
    batch_size = 2
    hidden_size = 4
    attention_heads = 2
    sequence_length = 3

    nodes = GenerateNodes2(attention_heads)

    initializers = GenerateInitializers2(hidden_size)

    graph = helper.make_graph(
        nodes,
        "EmbedLayerNorm_format7",  # name
        [  # inputs
            helper.make_tensor_value_info("input_ids", TensorProto.INT64, [batch_size, sequence_length]),
            helper.make_tensor_value_info("input_mask", TensorProto.INT64, [batch_size, sequence_length]),
        ],
        [  # outputs
            helper.make_tensor_value_info(
                "add3_out",
                TensorProto.FLOAT,
                [batch_size, sequence_length, hidden_size],
            ),
        ],
        initializers,
    )

    model = helper.make_model(graph)
    onnx.save(model, model_name)


def GenerateModel8(model_name):  # noqa: N802
    batch_size = -1
    hidden_size = 4
    attention_heads = 2
    sequence_length = -1

    nodes = GenerateNodes2(attention_heads)

    del nodes[5:7]
    del nodes[1:3]
    new_nodes = [
        helper.make_node("Shape", ["input_ids"], ["shape_out"], "shape"),
        helper.make_node("Gather", ["shape_out", "indices_1"], ["gather0_out"], "gather0"),
        helper.make_node("Expand", ["unsqueeze0_out", "shape_out"], ["expand_out"], "expand"),
    ]
    nodes = nodes + new_nodes

    initializers = GenerateInitializers2(hidden_size)

    graph = helper.make_graph(
        nodes,
        "EmbedLayerNorm_format8",  # name
        [  # inputs
            helper.make_tensor_value_info("input_ids", TensorProto.INT64, [batch_size, sequence_length]),
            helper.make_tensor_value_info("input_mask", TensorProto.INT64, [batch_size, sequence_length]),
        ],
        [  # outputs
            helper.make_tensor_value_info(
                "add3_out",
                TensorProto.FLOAT,
                [batch_size, sequence_length, hidden_size],
            ),
        ],
        initializers,
    )

    model = helper.make_model(graph)
    onnx.save(model, model_name)


def GenerateModel9(model_name):  # noqa: N802
    batch_size = -1
    hidden_size = 4
    attention_heads = 2
    sequence_length = -1

    nodes = GenerateNodes2(attention_heads)

    del nodes[10]
    del nodes[5:7]
    del nodes[1:3]
    new_nodes = [
        helper.make_node("Shape", ["input_ids"], ["shape_out"], "shape"),
        helper.make_node("Gather", ["shape_out", "indices_1"], ["gather0_out"], "gather0"),
        helper.make_node("Expand", ["unsqueeze0_out", "shape_out"], ["expand_out"], "expand"),
        helper.make_node("Gather", ["shape_out", "indices_0"], ["gather1_out"], "gather1"),
        helper.make_node("Gather", ["shape_out", "indices_1"], ["gather2_out"], "gather2"),
        (
            helper.make_node("Unsqueeze", ["gather1_out", "axes_0"], ["unsqueeze1_out"], "unsqueeze1")
            if opset_version == 13
            else helper.make_node("Unsqueeze", ["gather1_out"], ["unsqueeze1_out"], "unsqueeze1", axes=[0])
        ),
        (
            helper.make_node("Unsqueeze", ["gather2_out", "axes_0"], ["unsqueeze2_out"], "unsqueeze2")
            if opset_version == 13
            else helper.make_node("Unsqueeze", ["gather2_out"], ["unsqueeze2_out"], "unsqueeze2", axes=[0])
        ),
        helper.make_node(
            "Concat",
            ["unsqueeze1_out", "unsqueeze2_out"],
            ["concat_out"],
            "concat",
            axis=0,
        ),
        helper.make_node(
            "ConstantOfShape",
            ["concat_out"],
            ["constant_of_shape_out"],
            "constant_of_shape",
            value=helper.make_tensor("mask_shape", TensorProto.FLOAT, [1], [1.0]),
        ),
        helper.make_node("Cast", ["constant_of_shape_out"], ["mask_cast_out"], "mask_cast", to=6),
    ]
    nodes = nodes + new_nodes

    initializers = GenerateInitializers2(hidden_size)

    graph = helper.make_graph(
        nodes,
        "EmbedLayerNorm_format9",  # name
        [  # inputs
            helper.make_tensor_value_info("input_ids", TensorProto.INT64, [batch_size, sequence_length]),
        ],
        [  # outputs
            helper.make_tensor_value_info(
                "add3_out",
                TensorProto.FLOAT,
                [batch_size, sequence_length, hidden_size],
            ),
        ],
        initializers,
    )

    model = helper.make_model(graph)
    onnx.save(model, model_name)


if opset_version == 11:
    GenerateModel3("embed_layer_norm_format3.onnx", True)
    GenerateModel3("embed_layer_norm_format3_no_cast.onnx", False)
    GenerateModel5("embed_layer_norm_format5.onnx")
    GenerateModel6("embed_layer_norm_format6.onnx")
    GenerateModel7("embed_layer_norm_format7.onnx")  # distilbert
    GenerateModel8("embed_layer_norm_format8.onnx")  # distilbert & shape nodes integration with input mask
    GenerateModel9("embed_layer_norm_format9.onnx")  # distilbert & shape nodes integration without input mask
    GenerateMultipleEmbedModel("embed_layer_norm_multiple.onnx")
else:
    GenerateModel3("embed_layer_norm_format3_opset13.onnx", True)
    GenerateModel3("embed_layer_norm_format3_no_cast_opset13.onnx", False)
    GenerateModel5("embed_layer_norm_format5_opset13.onnx")
    GenerateModel6("embed_layer_norm_format6_opset13.onnx")
    GenerateModel7("embed_layer_norm_format7_opset13.onnx")
    GenerateModel8("embed_layer_norm_format8_opset13.onnx")
    GenerateModel9("embed_layer_norm_format9_opset13.onnx")
    GenerateMultipleEmbedModel("embed_layer_norm_multiple_opset13.onnx")



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License.
#
# Copyright 2020 The Microsoft DeepSpeed Team
# Copyright (c) 2020, NVIDIA CORPORATION.
# Some functions in this file are adapted from following sources:
# - has_overflow_serial : https://github.com/microsoft/DeepSpeed/blob/d8e9ef6f99e27bb95e10bd146d145b3372b4cfda/deepspeed/runtime/zero/stage2.py#L1792
# - get_grad_norm_direct : https://github.com/microsoft/DeepSpeed/blob/d8e9ef6f99e27bb95e10bd146d145b3372b4cfda/deepspeed/runtime/zero/stage2.py#L1466
# - has_overflow_partitioned_grads_serial : https://github.com/microsoft/DeepSpeed/blob/d8e9ef6f99e27bb95e10bd146d145b3372b4cfda/deepspeed/runtime/zero/stage2.py#L1799
# --------------------------------------------------------------------------

from __future__ import annotations

import inspect
import types
import warnings

import torch
from numpy import inf
from packaging.version import Version

from ._ds_code_store import Stage1And2_DeepSpeedZeroOptimizer_0_9_2
from ._modifier import FP16OptimizerModifier, check_overflow, check_overflow_for_grads
from ._multi_tensor_apply import MultiTensorApply

multi_tensor_applier = MultiTensorApply(2048 * 32)


def _get_normalized_str(function) -> str:
    return inspect.getsource(function)


def _dynamic_checks(cur_ds_version: Version, optimizer) -> bool:
    _functions_to_override = ["has_overflow_serial", "get_grad_norm_direct", "has_overflow_partitioned_grads_serial"]

    _version_to_source_code_map = {"0.9.2": Stage1And2_DeepSpeedZeroOptimizer_0_9_2}

    # Try to find the biggest version that is smaller than or equal to cur_ds_version.
    # then compare the source code (in case the found version is the latest version supported);
    # If current code does not match the found version, return False, and raise a warning to
    # add the new version to the list.
    versions = [Version(v) for v in _version_to_source_code_map]
    sorted_versions = sorted(versions, reverse=True)
    version_to_compare = None
    for sv in sorted_versions:
        if cur_ds_version >= sv:
            version_to_compare = sv
            break

    if version_to_compare is None:
        warnings.warn(
            "Unable to find a DeepSpeed version that is smaller than or equal to the current version "
            f"{cur_ds_version}. Skip modifying optimizer.",
            UserWarning,
        )
        return False

    v_optimizer_cls = _version_to_source_code_map[str(version_to_compare)]
    all_match = True
    for func_name in _functions_to_override:
        if not getattr(optimizer, func_name):
            warnings.warn(
                f"DeepSpeed function {func_name} is not found in optimizer. Skip modifying optimizer.", UserWarning
            )
            all_match = False
        cur_code_str = _get_normalized_str(getattr(optimizer, func_name))
        v_code_str = _get_normalized_str(getattr(v_optimizer_cls, func_name))
        if cur_code_str != v_code_str:
            warnings.warn(
                f"DeepSpeed function {func_name} has changed after version {version_to_compare}. "
                f"Please append new version {cur_ds_version} in _version_to_source_code_map and _ds_code_store.py.\n"
                f"---[{func_name}] Old Source Code Start----\n"
                f"{v_code_str}\n"
                f"---{func_name} Old Source Code End----\n"
                f"---[{func_name}] New Source Code Start----\n"
                f"{cur_code_str}\n"
                f"---{func_name} New Source Code End----",
                UserWarning,
            )
            all_match = False

    return all_match


class DeepSpeedZeROModifier(FP16OptimizerModifier):
    def __init__(self, optimizer, **kwargs) -> None:
        super().__init__(optimizer)

    def can_be_modified(self):
        import deepspeed

        # Note 1:
        # This modifier relies on the implementation of has_overflow_serial, get_grad_norm_direct,
        # and has_overflow_partitioned_grads_serial
        # in https://github.com/microsoft/DeepSpeed/blob/master/deepspeed/runtime/zero/stage_1_and_2.py.
        # The minimum version supported is 0.4.0, all versions in between [0.4.0, 0.9.1]
        # are manually checked to make sure the implementation of these functions are "logically" not changed.
        # The way we did the check is to check the history of this file, if there is no change during the update,
        # it's safe to update the version supporting list. Otherwise, or the file is moved or renamed,
        # we need to check the implementation of these functions in detail.
        #
        # Note 2:
        # Since version 0.9.2, we added dynamic source code check, by comparing installed version of code with
        # the source code in our code store. If the source code is changed, we will raise a warning to ask user
        # to add the new version to the code store. Otherwise, we will override the functions.

        ds_version = Version(deepspeed.__version__)
        if ds_version < Version("0.4.0"):
            warnings.warn(
                f"Skip modifying optimizer because of unsupported DeepSpeed version {ds_version}, "
                "minimum supported version: 0.4.0, current version",
                UserWarning,
            )
            return False
        if ds_version > Version("0.9.1") and not _dynamic_checks(ds_version, self._optimizer):
            warnings.warn(
                f"Skip modifying optimizer because of unsupported DeepSpeed version {ds_version}.",
                UserWarning,
            )
            return False

        try:
            from deepspeed.accelerator import get_accelerator
        except ImportError:
            warnings.warn("Unable to import get_accelerator from deepspeed.accelerator", UserWarning)
        else:
            if not get_accelerator().device_name().startswith("cuda"):
                warnings.warn(
                    "Skip modifying optimizer as device is not supported, "
                    f"device name: {get_accelerator().device_name()}",
                    UserWarning,
                )
                return False

        return self.check_requirements(
            ["has_overflow_serial", "get_grad_norm_direct", "has_overflow_partitioned_grads_serial"],
            require_apex=False,
            require_torch_non_finite_check=True,
        )

    def override_function(self):
        warnings.warn("DeepSpeed fp16_optimizer functions are overridden with faster implementation.", UserWarning)

        def get_grad_norm_direct(target, gradients, params, norm_type=2):
            from onnxruntime.training.ortmodule.torch_cpp_extensions import fused_ops

            def is_model_parallel_parameter(p):
                return hasattr(p, "model_parallel") and p.model_parallel

            norm_type = float(norm_type)
            if norm_type == inf:
                total_norm = max(g.data.abs().max() for g in gradients)
                total_norm_cuda = torch.cuda.FloatTensor([float(total_norm)])
                torch.distributed.all_reduce(
                    total_norm_cuda, op=torch.distributed.ReduceOp.MAX, group=target.dp_process_group
                )

                # Take max across all GPUs.
                target._model_parallel_all_reduce(tensor=total_norm_cuda, op=torch.distributed.ReduceOp.MAX)
                total_norm = total_norm_cuda[0].item()
            else:
                total_norm = 0.0

                #### THIS IS THE ORIGINAL IMPLEMENTATION ####
                # # if dist.get_rank() == 0:
                # #    logger.info(f"Total Norm beginning {total_norm}")
                # for g, p in zip(gradients, params):
                #     # Pipeline parallelism may replicate parameters. Avoid multi-counting.
                #     if hasattr(p, 'ds_pipe_replicated') and p.ds_pipe_replicated:
                #         continue
                #     if is_model_parallel_parameter(p) or (self.model_parallel_rank == 0):
                #         param_norm = g.data.double().norm(2)
                #         total_norm += param_norm.item()**2
                # # Sum across all model parallel GPUs.
                # total_norm_cuda = torch.cuda.FloatTensor([float(total_norm)])
                #### END OF THE ORIGINAL IMPLEMENTATION ####

                #### THIS IS THE FASTER IMPLEMENTATION ####
                grads_for_norm = []
                for g, p in zip(gradients, params, strict=False):
                    if is_model_parallel_parameter(p) or (target.model_parallel_rank == 0):
                        # BE NOTED: deepspeed original give a double type conversion here, not sure whether this is impacting some models.
                        # https://github.com/microsoft/DeepSpeed/blob/9e5c0c5c3ecabb68b7e9dffac0e9b8d167e3cab8/deepspeed/runtime/zero/stage2.py#L1501
                        # grads_for_norm.append(g.data.double())
                        grads_for_norm.append(g.data)

                if len(grads_for_norm) > 0:
                    dummy_overflow_buf = torch.cuda.IntTensor([0])
                    # Use apex's multi-tensor applier for efficiency reasons.
                    # Multi-tensor applier takes a function and a list of list
                    # and performs the operation on that list all in one kernel.
                    grad_norm, _ = multi_tensor_applier(
                        fused_ops.multi_tensor_l2norm,
                        dummy_overflow_buf,
                        [fused_ops.TorchTensorVector(grads_for_norm)],
                        False,  # no per-parameter norm
                    )
                    # Since we will be summing across data parallel groups,
                    # we need the pow(norm-type).
                    total_norm_cuda = grad_norm**norm_type
                else:
                    total_norm_cuda = torch.cuda.FloatTensor([float(total_norm)])
                #### END OF THE FASTER IMPLEMENTATION ####

                # Sum across all model parallel GPUs.
                torch.distributed.all_reduce(
                    total_norm_cuda, op=torch.distributed.ReduceOp.SUM, group=target.dp_process_group
                )

                target._model_parallel_all_reduce(tensor=total_norm_cuda, op=torch.distributed.ReduceOp.SUM)

                total_norm = total_norm_cuda[0].item() ** (1.0 / norm_type)

            if total_norm == float("inf") or total_norm == -float("inf") or total_norm != total_norm:
                total_norm = -1

            return total_norm

        def has_overflow_serial(target, params, is_grad_list=False):
            #### THIS IS THE ORIGINAL IMPLEMENTATION ####
            # for p in params:
            #     if p.grad is not None and self._has_inf_or_nan(p.grad.data):
            #         return True
            #
            # return False
            #### END OF THE ORIGINAL IMPLEMENTATION ####

            #### THIS IS THE FASTER IMPLEMENTATION ####
            return check_overflow(params)
            #### END OF THE FASTER IMPLEMENTATION ####

        def has_overflow_partitioned_grads_serial(target):
            #### THIS IS THE ORIGINAL IMPLEMENTATION ####
            # for i in range(len(self.fp16_groups)):
            #     for j, grad in enumerate(self.averaged_gradients[i]):
            #         if grad is not None and self._has_inf_or_nan(grad.data, j):
            #             return True
            # return False
            #### END OF THE ORIGINAL IMPLEMENTATION ####

            #### THIS IS THE FASTER IMPLEMENTATION ####
            groups = target.fp16_groups if hasattr(target, "fp16_groups") else target.bit16_groups
            for i in range(len(groups)):
                grad_data = [grad.data for grad in target.averaged_gradients[i] if grad is not None]
                if check_overflow_for_grads(grad_data):
                    return True
            return False
            #### END OF THE FASTER IMPLEMENTATION ####

        self._optimizer.has_overflow_serial = types.MethodType(has_overflow_serial, self._optimizer)
        self._optimizer.get_grad_norm_direct = types.MethodType(get_grad_norm_direct, self._optimizer)
        # zero1 should not call into following function, is this a deepspeed bug?
        self._optimizer.has_overflow_partitioned_grads_serial = types.MethodType(
            has_overflow_partitioned_grads_serial, self._optimizer
        )
